# Copyright 2014-2015, Tresys Technology, LLC
# Copyright 2017, Chris PeBenito <pebenito@ieee.org>
#
# This file is part of SETools.
#
# SETools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1 of
# the License, or (at your option) any later version.
#
# SETools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with SETools.  If not, see
# <http://www.gnu.org/licenses/>.
#
import ipaddress

import logging
from socket import AF_INET, AF_INET6
from typing import Iterable, Optional, Union

from .mixins import MatchContext
from .policyrep import Nodecon, NodeconIPVersion
from .query import PolicyQuery

AnyIPNetwork = Union[ipaddress.IPv4Network, ipaddress.IPv6Network]


class NodeconQuery(MatchContext, PolicyQuery):

    """
    Query nodecon statements.

    Parameter:
    policy          The policy to query.

    Keyword Parameters/Class attributes:
    network         The IPv4/IPv6 address or IPv4/IPv6 network address
                    with netmask, e.g. 192.168.1.0/255.255.255.0 or
                    "192.168.1.0/24".
    network_overlap If true, the net will match if it overlaps with
                    the nodecon's network instead of equality.
    ip_version      The IP version of the nodecon to match. (socket.AF_INET
                    for IPv4 or socket.AF_INET6 for IPv6)
    user            The criteria to match the context's user.
    user_regex      If true, regular expression matching
                    will be used on the user.
    role            The criteria to match the context's role.
    role_regex      If true, regular expression matching
                    will be used on the role.
    type_           The criteria to match the context's type.
    type_regex      If true, regular expression matching
                    will be used on the type.
    range_          The criteria to match the context's range.
    range_subset    If true, the criteria will match if it is a subset
                    of the context's range.
    range_overlap   If true, the criteria will match if it overlaps
                    any of the context's range.
    range_superset  If true, the criteria will match if it is a superset
                    of the context's range.
    range_proper    If true, use proper superset/subset operations.
                    No effect if not using set operations.
    """

    _network: Optional[AnyIPNetwork] = None
    network_overlap: bool = False
    _ip_version: Optional[NodeconIPVersion] = None

    @property
    def ip_version(self) -> Optional[NodeconIPVersion]:
        return self._ip_version

    @ip_version.setter
    def ip_version(self, value: Optional[Union[str, NodeconIPVersion]]) -> None:
        if value:
            self._ip_version = NodeconIPVersion.lookup(value)
        else:
            self._ip_version = None

    @property
    def network(self) -> Optional[AnyIPNetwork]:
        return self._network

    @network.setter
    def network(self, value: Optional[Union[str, AnyIPNetwork]]) -> None:
        if value:
            self._network = ipaddress.ip_network(value)
        else:
            self._network = None

    def __init__(self, policy, **kwargs) -> None:
        super(NodeconQuery, self).__init__(policy, **kwargs)
        self.log = logging.getLogger(__name__)

    def results(self) -> Iterable[Nodecon]:
        """Generator which yields all matching nodecons."""
        self.log.info("Generating nodecon results from {0.policy}".format(self))
        self.log.debug("Network: {0.network!r}, overlap: {0.network_overlap}".format(self))
        self.log.debug("IP Version: {0.ip_version!r}".format(self))
        self._match_context_debug(self.log)

        for nodecon in self.policy.nodecons():

            if self.network:
                if self.network_overlap:
                    if not self.network.overlaps(nodecon.network):  # type: ignore
                        continue
                else:
                    if not nodecon.network == self.network:
                        continue

            if self.ip_version and self.ip_version != nodecon.ip_version:
                continue

            if not self._match_context(nodecon.context):
                continue

            yield nodecon



# Copyright 2016, Tresys Technology, LLC
#
# This file is part of SETools.
#
# SETools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1 of
# the License, or (at your option) any later version.
#
# SETools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with SETools.  If not, see
# <http://www.gnu.org/licenses/>.
#
from PyQt5.QtGui import QCursor
from PyQt5.QtWidgets import QAction, QListView, QMenu


class GetDetailsListView(QListView):

    """A QListView widget with more details context menu."""

    def __init__(self, parent):
        super(GetDetailsListView, self).__init__(parent)

        # set up right-click context menu
        self.get_detail = QAction("More details...", self)
        self.menu = QMenu(self)
        self.menu.addAction(self.get_detail)

    def contextMenuEvent(self, event):
        self.menu.popup(QCursor.pos())



# Copyright 2016, Tresys Technology, LLC
#
# This file is part of SETools.
#
# SETools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1 of
# the License, or (at your option) any later version.
#
# SETools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with SETools.  If not, see
# <http://www.gnu.org/licenses/>.
#
from PyQt5.QtCore import pyqtSignal, QObject, QThread


class QueryResultsUpdater(QObject):

    """
    Thread for processing basic queries and updating result widgets.

    Parameters:
    query       The query object
    model       The model for the results

    Qt signals:
    finished    (int) The update has completed, with the number of results.
    raw_line    (str) A string to be appended to the raw results.
    """

    finished = pyqtSignal(int)
    raw_line = pyqtSignal(str)

    def __init__(self, query, model):
        super(QueryResultsUpdater, self).__init__()
        self.query = query
        self.model = model

    def update(self):
        """Run the query and update results."""
        self.model.beginResetModel()

        results = []
        counter = 0

        for counter, item in enumerate(self.query.results(), start=1):
            results.append(item)

            self.raw_line.emit(str(item))

            if QThread.currentThread().isInterruptionRequested():
                break
            elif not counter % 10:
                # yield execution every 10 rules
                QThread.yieldCurrentThread()

        self.model.resultlist = results
        self.model.endResetModel()

        self.finished.emit(counter)



# Copyright 2016, Tresys Technology, LLC
#
# This file is part of SETools.
#
# SETools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1 of
# the License, or (at your option) any later version.
#
# SETools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with SETools.  If not, see
# <http://www.gnu.org/licenses/>.
#
from typing import NamedTuple, Optional

from ..policyrep import Default, DefaultRuletype, DefaultValue, DefaultRangeValue, ObjClass

from .descriptors import DiffResultDescriptor
from .difference import Difference, SymbolWrapper, Wrapper


class ModifiedDefault(NamedTuple):

    """Difference details for a modified default_*."""

    rule: Default
    added_default: Optional[DefaultValue]
    removed_default: Optional[DefaultValue]
    added_default_range: Optional[DefaultRangeValue]
    removed_default_range: Optional[DefaultRangeValue]


class DefaultsDifference(Difference):

    """Determine the difference in default_* between two policies."""

    added_defaults = DiffResultDescriptor("diff_defaults")
    removed_defaults = DiffResultDescriptor("diff_defaults")
    modified_defaults = DiffResultDescriptor("diff_defaults")

    def diff_defaults(self) -> None:
        """Generate the difference in type defaults between the policies."""

        self.log.info(
            "Generating default_* differences from {0.left_policy} to {0.right_policy}".
            format(self))

        self.added_defaults, self.removed_defaults, matched_defaults = self._set_diff(
            (DefaultWrapper(d) for d in self.left_policy.defaults()),
            (DefaultWrapper(d) for d in self.right_policy.defaults()))

        self.modified_defaults = []

        for left_default, right_default in matched_defaults:
            # Criteria for modified defaults
            # 1. change to default setting
            # 2. change to default range

            if left_default.default != right_default.default:
                removed_default = left_default.default
                added_default = right_default.default
            else:
                removed_default = None
                added_default = None

            try:
                if left_default.default_range != right_default.default_range:
                    removed_default_range = left_default.default_range
                    added_default_range = right_default.default_range
                else:
                    removed_default_range = None
                    added_default_range = None
            except AttributeError:
                removed_default_range = None
                added_default_range = None

            if removed_default or removed_default_range:
                self.modified_defaults.append(
                    ModifiedDefault(left_default,
                                    added_default,
                                    removed_default,
                                    added_default_range,
                                    removed_default_range))

    #
    # Internal functions
    #
    def _reset_diff(self) -> None:
        """Reset diff results on policy changes."""
        self.log.debug("Resetting default_* differences")
        self.added_defaults = None
        self.removed_defaults = None
        self.modified_defaults = None


# Pylint bug: https://github.com/PyCQA/pylint/issues/2822
class DefaultWrapper(Wrapper[Default]):  # pylint: disable=unsubscriptable-object

    """Wrap default_* to allow comparisons."""

    __slots__ = ("ruletype", "tclass")

    def __init__(self, default: Default) -> None:
        self.origin = default
        self.ruletype = default.ruletype
        self.tclass = SymbolWrapper(default.tclass)
        self.key = hash(default)

    def __hash__(self):
        return self.key

    def __lt__(self, other):
        return self.key < other.key

    def __eq__(self, other):
        return self.ruletype == other.ruletype and \
            self.tclass == other.tclass



# Copyright 2014-2015, Tresys Technology, LLC
#
# This file is part of SETools.
#
# SETools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1 of
# the License, or (at your option) any later version.
#
# SETools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with SETools.  If not, see
# <http://www.gnu.org/licenses/>.
#
import logging
from typing import cast, Iterable, Optional, Set, Tuple

from . import mixins, query
from .descriptors import CriteriaDescriptor, CriteriaSetDescriptor
from .exception import RuleUseError, RuleNotConditional
from .policyrep import AnyTERule, AVRuleXperm, IoctlSet, TERuletype
from .util import match_indirect_regex, match_regex_or_set


class TERuleQuery(mixins.MatchObjClass, mixins.MatchPermission, query.PolicyQuery):

    """
    Query the Type Enforcement rules.

    Parameter:
    policy            The policy to query.

    Keyword Parameters/Class attributes:
    ruletype          The list of rule type(s) to match.
    source            The name of the source type/attribute to match.
    source_indirect   If true, members of an attribute will be
                      matched rather than the attribute itself.
                      Default is true.
    source_regex      If true, regular expression matching will
                      be used on the source type/attribute.
                      Obeys the source_indirect option.
                      Default is false.
    target            The name of the target type/attribute to match.
    target_indirect   If true, members of an attribute will be
                      matched rather than the attribute itself.
                      Default is true.
    target_regex      If true, regular expression matching will
                      be used on the target type/attribute.
                      Obeys target_indirect option.
                      Default is false.
    tclass            The object class(es) to match.
    tclass_regex      If true, use a regular expression for
                      matching the rule's object class.
                      Default is false.
    perms             The set of permission(s) to match.
    perms_equal       If true, the permission set of the rule
                      must exactly match the permissions
                      criteria.  If false, any set intersection
                      will match.
                      Default is false.
    perms_regex       If true, regular expression matching will be used
                      on the permission names instead of set logic.
                      Default is false.
    perms_subset      If true, the rule matches if the permissions criteria
                      is a subset of the rule's permission set.
                      Default is false.
    default           The name of the default type to match.
    default_regex     If true, regular expression matching will be
                      used on the default type.
                      Default is false.
    boolean           The set of boolean(s) to match.
    boolean_regex     If true, regular expression matching will be
                      used on the booleans.
                      Default is false.
    boolean_equal     If true, the booleans in the conditional
                      expression of the rule must exactly match the
                      criteria.  If false, any set intersection
                      will match.  Default is false.
    """

    ruletype = CriteriaSetDescriptor(enum_class=TERuletype)
    source = CriteriaDescriptor("source_regex", "lookup_type_or_attr")
    source_regex: bool = False
    source_indirect: bool = True
    target = CriteriaDescriptor("target_regex", "lookup_type_or_attr")
    target_regex: bool = False
    target_indirect: bool = True
    default = CriteriaDescriptor("default_regex", "lookup_type_or_attr")
    default_regex: bool = False
    boolean = CriteriaSetDescriptor("boolean_regex", "lookup_boolean")
    boolean_regex: bool = False
    boolean_equal: bool = False
    _xperms: Optional[IoctlSet] = None
    xperms_equal: bool = False

    @property
    def xperms(self) -> Optional[IoctlSet]:
        return self._xperms

    @xperms.setter
    def xperms(self, value: Optional[Iterable[Tuple[int, int]]]) -> None:
        if value:
            pending_xperms: Set[int] = set()

            for low, high in value:
                if not (0 <= low <= 0xffff):
                    raise ValueError("{0:#07x} is not a valid ioctl.".format(low))

                if not (0 <= high <= 0xffff):
                    raise ValueError("{0:#07x} is not a valid ioctl.".format(high))

                if high < low:
                    high, low = low, high

                pending_xperms.update(i for i in range(low, high + 1))

            self._xperms = IoctlSet(pending_xperms)
        else:
            self._xperms = None

    def __init__(self, policy, **kwargs) -> None:
        super(TERuleQuery, self).__init__(policy, **kwargs)
        self.log = logging.getLogger(__name__)

    def results(self) -> Iterable[AnyTERule]:
        """Generator which yields all matching TE rules."""
        self.log.info("Generating TE rule results from {0.policy}".format(self))
        self.log.debug("Ruletypes: {0.ruletype}".format(self))
        self.log.debug("Source: {0.source!r}, indirect: {0.source_indirect}, "
                       "regex: {0.source_regex}".format(self))
        self.log.debug("Target: {0.target!r}, indirect: {0.target_indirect}, "
                       "regex: {0.target_regex}".format(self))
        self._match_object_class_debug(self.log)
        self._match_perms_debug(self.log)
        self.log.debug("Xperms: {0.xperms!r}, eq: {0.xperms_equal}".format(self))
        self.log.debug("Default: {0.default!r}, regex: {0.default_regex}".format(self))
        self.log.debug("Boolean: {0.boolean!r}, eq: {0.boolean_equal}, "
                       "regex: {0.boolean_regex}".format(self))

        for rule in self.policy.terules():
            #
            # Matching on rule type
            #
            if self.ruletype:
                if rule.ruletype not in self.ruletype:
                    continue

            #
            # Matching on source type
            #
            if self.source and not match_indirect_regex(
                    rule.source,
                    self.source,
                    self.source_indirect,
                    self.source_regex):
                continue

            #
            # Matching on target type
            #
            if self.target and not match_indirect_regex(
                    rule.target,
                    self.target,
                    self.target_indirect,
                    self.target_regex):
                continue

            #
            # Matching on object class
            #
            if not self._match_object_class(rule):
                continue

            #
            # Matching on permission set
            #
            try:
                if self.perms and rule.extended:
                    if self.perms_equal and len(self.perms) > 1:
                        # if criteria is more than one standard permission,
                        # extended perm rules can never match if the
                        # permission set equality option is on.
                        continue

                    if cast(AVRuleXperm, rule).xperm_type not in self.perms:
                        continue
                elif not self._match_perms(rule):
                    continue
            except RuleUseError:
                continue

            #
            # Matching on extended permissions
            #
            try:
                if self.xperms and not match_regex_or_set(
                        rule.perms,
                        self.xperms,
                        self.xperms_equal,
                        False):
                    continue

            except RuleUseError:
                continue

            #
            # Matching on default type
            #
            if self.default:
                try:
                    # because default type is always a single
                    # type, hard-code indirect to True
                    # so the criteria can be an attribute
                    if not match_indirect_regex(
                            rule.default,
                            self.default,
                            True,
                            self.default_regex):
                        continue
                except RuleUseError:
                    continue

            #
            # Match on Boolean in conditional expression
            #
            if self.boolean:
                try:
                    if not match_regex_or_set(
                            rule.conditional.booleans,
                            self.boolean,
                            self.boolean_equal,
                            self.boolean_regex):
                        continue
                except RuleNotConditional:
                    continue

            # if we get here, we have matched all available criteria
            yield rule



# Copyright 2020, Microsoft Corporation
#
# This file is part of SETools.
#
# SETools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1 of
# the License, or (at your option) any later version.
#
# SETools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with SETools.  If not, see
# <http://www.gnu.org/licenses/>.
#

import sys
import logging
from abc import ABCMeta, abstractmethod
from typing import Dict, FrozenSet, List, Mapping

from ..exception import InvalidCheckOption
from ..policyrep import SELinuxPolicy

from .globalkeys import CHECK_TYPE_KEY, CHECK_DESC_KEY, CHECK_DISABLE, GLOBAL_CONFIG_KEYS


CHECKER_REGISTRY: Dict[str, type] = {}

__all__ = ['CheckerModule']


class CheckRegistry(ABCMeta):

    """Checker module registry metaclass.  This registers modules in the check registry."""

    def __new__(cls, clsname, superclasses, attributedict):
        check_type = attributedict.get("check_type")
        check_config = attributedict.get("check_config")

        if clsname != "CheckerModule":
            if not isinstance(check_type, str):
                raise TypeError("Checker module {} does not set a check_type.".format(clsname))

            if not isinstance(check_config, frozenset):
                raise TypeError("Checker module {} does not set a valid check_config.".format(
                    clsname))

            if check_type in CHECKER_REGISTRY:
                existing_check_module = CHECKER_REGISTRY[check_type].__name__
                raise TypeError("Checker module {} conflicts with registered check {}".format(
                                clsname, existing_check_module))

        classdef = super().__new__(cls, clsname, superclasses, attributedict)

        if check_type:
            CHECKER_REGISTRY[check_type] = classdef

        return classdef


class CheckerModule(metaclass=CheckRegistry):

    """Abstract base class for policy checker modules."""

    # The name of the check used in config files.
    # This must be set by subclasses.
    check_type: str

    # The container of valid config keys specific to the check
    # This is in addition to the common config keys
    # in the GLOBAL_CONFIG_KEYS above.  This must be set by subclasses.
    # If no additional keys  are needed, this should be set to an
    # empty container.
    check_config: FrozenSet[str]

    # T/F log findings that pass the check.
    log_passing: bool = False

    # Default output to stdout.
    output = sys.stdout

    policy: SELinuxPolicy

    def __init__(self, policy: SELinuxPolicy, checkname: str, config: Mapping[str, str]) -> None:
        self.policy = policy
        self.checkname = checkname

        # ensure there is a logger available. This should
        # be replaced with the concrete class' logger
        self.log = logging.getLogger(__name__)

        # Check available options are valid
        valid_options = GLOBAL_CONFIG_KEYS | self.check_config
        for k in config:
            if k not in valid_options:
                raise InvalidCheckOption("{}: Invalid option: {}".format(
                    self.checkname, k))

        # Make sure all global config attrs are initialized for this check
        self.desc = config.get(CHECK_DESC_KEY)
        self.disable = config.get(CHECK_DISABLE)

    def log_info(self, msg: str) -> None:
        """Output an informational message."""
        self.output.write(msg)
        self.output.write("\n")
        self.log.debug(msg)

    def log_ok(self, msg: str) -> None:
        """
        Log findings that pass the check.  By default these messages are
        surpressed unless self.log_passing is True.
        """
        if self.log_passing:
            self.output.write("P   * {}\n".format(msg))
        self.log.debug("P   * {}".format(msg))

    def log_fail(self, msg: str) -> None:
        """Log findings that fail the check."""
        self.output.write("{}   * {}\n".format("F" if self.log_passing else " ", msg))
        self.log.debug("F   * {}".format(msg))

    @abstractmethod
    def run(self) -> List:
        """
        Run the configured check on the policy.

        Return:   List of failed items in the check.  If the check passes, list is empty.
        """
        pass



# Copyright 2015-2016, Tresys Technology, LLC
#
# This file is part of SETools.
#
# SETools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1 of
# the License, or (at your option) any later version.
#
# SETools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with SETools.  If not, see
# <http://www.gnu.org/licenses/>.
#
from .bool import BooleansDifference
from .bounds import BoundsDifference
from .commons import CommonDifference
from .constraints import ConstraintsDifference
from .default import DefaultsDifference
from .fsuse import FSUsesDifference
from .genfscon import GenfsconsDifference
from .ibendportcon import IbendportconsDifference
from .ibpkeycon import IbpkeyconsDifference
from .initsid import InitialSIDsDifference
from .mls import CategoriesDifference, LevelDeclsDifference, SensitivitiesDifference
from .mlsrules import MLSRulesDifference
from .netifcon import NetifconsDifference
from .nodecon import NodeconsDifference
from .objclass import ObjClassDifference
from .polcap import PolCapsDifference
from .portcon import PortconsDifference
from .properties import PropertiesDifference
from .rbacrules import RBACRulesDifference
from .roles import RolesDifference
from .terules import TERulesDifference
from .typeattr import TypeAttributesDifference
from .types import TypesDifference
from .users import UsersDifference

__all__ = ['PolicyDifference']


class PolicyDifference(BooleansDifference,
                       BoundsDifference,
                       CategoriesDifference,
                       CommonDifference,
                       ConstraintsDifference,
                       DefaultsDifference,
                       FSUsesDifference,
                       GenfsconsDifference,
                       IbendportconsDifference,
                       IbpkeyconsDifference,
                       InitialSIDsDifference,
                       LevelDeclsDifference,
                       MLSRulesDifference,
                       NetifconsDifference,
                       NodeconsDifference,
                       ObjClassDifference,
                       PolCapsDifference,
                       PortconsDifference,
                       PropertiesDifference,
                       RBACRulesDifference,
                       RolesDifference,
                       SensitivitiesDifference,
                       TERulesDifference,
                       TypeAttributesDifference,
                       TypesDifference,
                       UsersDifference):

    """
    Determine the differences from the left policy to the right policy.

    Parameters:
    left    A policy
    right   A policy
    """

    def _reset_diff(self):
        """Reset diff results on policy changes."""
        for c in PolicyDifference.__bases__:
            c._reset_diff(self)



# Copyright 2016, Tresys Technology, LLC
#
# This file is part of SETools.
#
# SETools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1 of
# the License, or (at your option) any later version.
#
# SETools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with SETools.  If not, see
# <http://www.gnu.org/licenses/>.
#
from PyQt5.QtCore import Qt
from setools.exception import ConstraintUseError

from .models import SEToolsTableModel


class ConstraintTableModel(SEToolsTableModel):

    """A table-based model for constraints."""

    headers = ["Rule Type", "Class", "Permissions", "Expression"]

    def data(self, index, role):
        if self.resultlist and index.isValid():
            row = index.row()
            col = index.column()
            rule = self.resultlist[row]

            if role == Qt.DisplayRole:
                if col == 0:
                    return rule.ruletype.name
                elif col == 1:
                    return rule.tclass.name
                elif col == 2:
                    try:
                        return ", ".join(sorted(rule.perms))
                    except ConstraintUseError:
                        return None
                elif col == 3:
                    return str(rule.expression)

            elif role == Qt.UserRole:
                return rule



# Copyright 2016, Tresys Technology, LLC
#
# This file is part of SETools.
#
# SETools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1 of
# the License, or (at your option) any later version.
#
# SETools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with SETools.  If not, see
# <http://www.gnu.org/licenses/>.
#
from collections import defaultdict

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QDialog, QTreeWidgetItem

from ..widget import SEToolsWidget
from .analysistab import AnalysisSection, AnalysisTab, TAB_REGISTRY


class ChooseAnalysis(SEToolsWidget, QDialog):

    """
    Dialog for choosing a new analysis

    The below class attributes are used for populating
    the GUI contents and mapping them to the appropriate
    tab widget class for the analysis.
    """

    def __init__(self, parent):
        super(ChooseAnalysis, self).__init__(parent)
        self.parent = parent

        # populate the analysis choices tree:
        self.analysis_choices = defaultdict(dict)
        for clsobj in TAB_REGISTRY.values():
            self.analysis_choices[clsobj.section.name][clsobj.tab_title] = clsobj

        self.setupUi()

    def setupUi(self):
        self.load_ui("apol/choose_analysis.ui")

    def show(self, mls):
        self.analysisTypes.clear()
        for groupname, group in self.analysis_choices.items():
            groupitem = QTreeWidgetItem(self.analysisTypes)
            groupitem.setText(0, groupname)
            groupitem._tab_class = None
            for entryname, cls in group.items():
                if cls.mlsonly and not mls:
                    continue

                item = QTreeWidgetItem(groupitem)
                item.setText(0, entryname)
                item._tab_class = cls
                groupitem.addChild(item)

        self.analysisTypes.expandAll()
        self.analysisTypes.sortByColumn(0, Qt.AscendingOrder)
        super(ChooseAnalysis, self).show()

    def accept(self, item=None):
        try:
            if not item:
                # .ui is set for single item selection.
                item = self.analysisTypes.selectedItems()[0]

            title = item.text(0)
            self.parent.create_new_analysis(title, item._tab_class)
        except (IndexError, TypeError):
            # IndexError: nothing is selected
            # TypeError: one of the group items was selected.
            pass
        else:
            super(ChooseAnalysis, self).accept()



# Copyright 2016, Tresys Technology, LLC
#
# This file is part of SETools.
#
# SETools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1 of
# the License, or (at your option) any later version.
#
# SETools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with SETools.  If not, see
# <http://www.gnu.org/licenses/>.
#

import logging
from contextlib import suppress

from PyQt5.QtCore import Qt, QSortFilterProxyModel, QStringListModel, QThread
from PyQt5.QtGui import QPalette, QTextCursor
from PyQt5.QtWidgets import QCompleter, QHeaderView, QMessageBox, QProgressDialog
from setools import ObjClassQuery

from ..logtosignal import LogHandlerToSignal
from ..models import PermListModel, SEToolsListModel, invert_list_selection
from ..objclassmodel import ObjClassTableModel, class_detail
from .analysistab import AnalysisSection, AnalysisTab
from .exception import TabFieldError
from .queryupdater import QueryResultsUpdater
from .workspace import load_checkboxes, load_lineedits, load_listviews, load_textedits, \
    save_checkboxes, save_lineedits, save_listviews, save_textedits


class ObjClassQueryTab(AnalysisTab):

    """ObjClass browser and query tab."""

    section = AnalysisSection.Components
    tab_title = "Object Classes"
    mlsonly = False

    def __init__(self, parent, policy, perm_map):
        super(ObjClassQueryTab, self).__init__(parent)
        self.log = logging.getLogger(__name__)
        self.policy = policy
        self.query = ObjClassQuery(policy)
        self.setupUi()

    def __del__(self):
        with suppress(RuntimeError):
            self.thread.quit()
            self.thread.wait(5000)

        logging.getLogger("setools.objclassquery").removeHandler(self.handler)

    def setupUi(self):
        self.load_ui("apol/objclassquery.ui")

        # populate class list
        self.class_model = SEToolsListModel(self)
        self.class_model.item_list = sorted(c for c in self.policy.classes())
        self.classes.setModel(self.class_model)

        # populate commons list
        self.common_model = SEToolsListModel(self)
        self.common_model.item_list = sorted(c for c in self.policy.commons())
        self.common.setModel(self.common_model)

        # populate perm list
        self.perms_model = PermListModel(self, self.policy)
        self.perms.setModel(self.perms_model)

        # set up results
        self.table_results_model = ObjClassTableModel(self)
        self.sort_proxy = QSortFilterProxyModel(self)
        self.sort_proxy.setSourceModel(self.table_results_model)
        self.table_results.setModel(self.sort_proxy)
        self.table_results.sortByColumn(0, Qt.AscendingOrder)

        # setup indications of errors on level/range
        self.errors = set()
        self.orig_palette = self.name.palette()
        self.error_palette = self.name.palette()
        self.error_palette.setColor(QPalette.Base, Qt.red)
        self.clear_name_error()

        # set up processing thread
        self.thread = QThread()
        self.worker = QueryResultsUpdater(self.query, self.table_results_model)
        self.worker.moveToThread(self.thread)
        self.worker.raw_line.connect(self.raw_results.appendPlainText)
        self.worker.finished.connect(self.update_complete)
        self.worker.finished.connect(self.thread.quit)
        self.thread.started.connect(self.worker.update)

        # create a "busy, please wait" dialog
        self.busy = QProgressDialog(self)
        self.busy.setModal(True)
        self.busy.setRange(0, 0)
        self.busy.setMinimumDuration(0)
        self.busy.canceled.connect(self.thread.requestInterruption)
        self.busy.reset()

        # update busy dialog from query INFO logs
        self.handler = LogHandlerToSignal()
        self.handler.message.connect(self.busy.setLabelText)
        logging.getLogger("setools.objclassquery").addHandler(self.handler)

        # Ensure settings are consistent with the initial .ui state
        self.notes.setHidden(not self.notes_expander.isChecked())

        # connect signals
        self.classes.doubleClicked.connect(self.get_detail)
        self.classes.get_detail.triggered.connect(self.get_detail)
        self.name.textEdited.connect(self.clear_name_error)
        self.name.editingFinished.connect(self.set_name)
        self.name_regex.toggled.connect(self.set_name_regex)
        self.common.selectionModel().selectionChanged.connect(self.set_common)
        self.perms.selectionModel().selectionChanged.connect(self.set_perms)
        self.invert_perms.clicked.connect(self.invert_perms_selection)
        self.buttonBox.clicked.connect(self.run)

    #
    # Class browser
    #
    def get_detail(self):
        # .ui is set for single item selection.
        index = self.classes.selectedIndexes()[0]
        item = self.class_model.data(index, Qt.UserRole)

        self.log.debug("Generating detail window for {0}".format(item))
        class_detail(self, item)

    #
    # Name criteria
    #
    def clear_name_error(self):
        self.clear_criteria_error(self.name, "Match the object class name.")

    def set_name(self):
        try:
            self.query.name = self.name.text()
        except Exception as ex:
            self.log.error("Object class name error: {0}".format(ex))
            self.set_criteria_error(self.name, ex)

    def set_name_regex(self, state):
        self.log.debug("Setting name_regex {0}".format(state))
        self.query.name_regex = state
        self.clear_name_error()
        self.set_name()

    #
    # Common criteria
    #
    def set_common(self):
        for index in self.common.selectionModel().selectedIndexes():
            self.query.common = self.common_model.data(index, Qt.UserRole)
            break
        else:
            self.query.common = None

    #
    # Permissions criteria
    #
    def set_perms(self):
        selected_perms = []
        for index in self.perms.selectionModel().selectedIndexes():
            selected_perms.append(self.perms_model.data(index, Qt.UserRole))

        self.query.perms = selected_perms

    def invert_perms_selection(self):
        invert_list_selection(self.perms.selectionModel())

    #
    # Save/Load tab
    #
    def save(self):
        """Return a dictionary of settings."""
        if self.errors:
            raise TabFieldError("Field(s) are in error: {0}".
                                format(" ".join(o.objectName() for o in self.errors)))

        settings = {}
        save_checkboxes(self, settings, ["criteria_expander", "notes_expander", "name_regex",
                                         "perms_equal"])
        save_lineedits(self, settings, ["name"])
        save_listviews(self, settings, ["common", "perms"])
        save_textedits(self, settings, ["notes"])
        return settings

    def load(self, settings):
        load_checkboxes(self, settings, ["criteria_expander", "notes_expander", "name_regex",
                                         "perms_equal"])
        load_lineedits(self, settings, ["name"])
        load_listviews(self, settings, ["common", "perms"])
        load_textedits(self, settings, ["notes"])

    #
    # Results runner
    #

    def run(self, button):
        # right now there is only one button.
        self.query.perms_equal = self.perms_equal.isChecked()

        # start processing
        self.busy.setLabelText("Processing query...")
        self.busy.show()
        self.raw_results.clear()
        self.thread.start()

    def update_complete(self, count):
        self.log.info("{0} object class(es) found.".format(count))

        # update sizes/location of result displays
        if not self.busy.wasCanceled():
            self.busy.setLabelText("Resizing the result table's columns; GUI may be unresponsive")
            self.busy.repaint()
            self.table_results.resizeColumnsToContents()
            # If the permissions column width is too long, pull back
            # to a reasonable size
            header = self.table_results.horizontalHeader()
            if header.sectionSize(1) > 400:
                header.resizeSection(1, 400)

        if not self.busy.wasCanceled():
            self.busy.setLabelText("Resizing the result table's rows; GUI may be unresponsive")
            self.busy.repaint()
            self.table_results.resizeRowsToContents()

        if not self.busy.wasCanceled():
            self.busy.setLabelText("Moving the raw result to top; GUI may be unresponsive")
            self.busy.repaint()
            self.raw_results.moveCursor(QTextCursor.Start)

        self.busy.reset()



# Copyright 2015, Tresys Technology, LLC
#
# This file is part of SETools.
#
# SETools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1 of
# the License, or (at your option) any later version.
#
# SETools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with SETools.  If not, see
# <http://www.gnu.org/licenses/>.
#
import sys
from errno import ENOENT

import pkg_resources
from PyQt5.uic import loadUi


# Stylesheet that adds a frame around QGroupBoxes
stylesheet = "\
QGroupBox {\
    border: 1px solid lightgrey;\
    margin-top: 0.5em;\
    }\
\
QGroupBox::title {\
    subcontrol-origin: margin;\
    left: 10px;\
    padding: 0 3px 0 3px;\
}\
"


class SEToolsWidget:
    def load_ui(self, filename):
        distro = pkg_resources.get_distribution("setools")
        path = "{0}/setoolsgui/{1}".format(distro.location, filename)
        loadUi(path, self)

        self.setStyleSheet(stylesheet)



# Copyright 2014-2015, Tresys Technology, LLC
#
# This file is part of SETools.
#
# SETools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1 of
# the License, or (at your option) any later version.
#
# SETools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with SETools.  If not, see
# <http://www.gnu.org/licenses/>.
#
import logging
from typing import Iterable

from .mixins import MatchName
from .policyrep import PolicyCapability
from .query import PolicyQuery


class PolCapQuery(MatchName, PolicyQuery):

    """
    Query SELinux policy capabilities

    Parameter:
    policy      The policy to query.

    Keyword Parameters/Class attributes:
    name        The name of the policy capability to match.
    name_regex  If true, regular expression matching will
                be used for matching the name.
    """

    def __init__(self, policy, **kwargs) -> None:
        super(PolCapQuery, self).__init__(policy, **kwargs)
        self.log = logging.getLogger(__name__)

    def results(self) -> Iterable[PolicyCapability]:
        """Generator which yields all matching policy capabilities."""
        self.log.info("Generating policy capability results from {0.policy}".format(self))
        self._match_name_debug(self.log)

        for cap in self.policy.polcaps():
            if not self._match_name(cap):
                continue

            yield cap



# Copyright 2016, Tresys Technology, LLC
#
# This file is part of SETools.
#
# SETools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1 of
# the License, or (at your option) any later version.
#
# SETools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with SETools.  If not, see
# <http://www.gnu.org/licenses/>.
#
import re
import logging

import setools


from PyQt5.QtCore import Qt, QItemSelectionModel


def save_checkboxes(tab, settings, checkboxes):
    """
    Save settings from the checkable buttons (e.g. QCheckbox) in the tab.

    Parameters:
    tab         The tab object.
    settings    The dictionary of settings.  This will be mutated.
    checkboxes  A list of attribute names (str) of buttons in the tab.
    """

    for entry in checkboxes:
        checkbox = getattr(tab, entry)
        settings[entry] = checkbox.isChecked()


def load_checkboxes(tab, settings, checkboxes):
    """
    Load settings into the checkable buttons (e.g. QCheckbox) in the tab.

    Parameters:
    tab         The tab object.
    settings    The dictionary of settings.
    checkboxes  A list of attribute names (str) of buttons in the tab.
    """

    log = logging.getLogger(__name__)

    # next set options
    for entry in checkboxes:
        checkbox = getattr(tab, entry)

        try:
            checkbox.setChecked(bool(settings[entry]))
        except KeyError:
            log.warning("{0} option missing from settings file.".format(entry))


def save_lineedits(tab, settings, lines):
    """
    Save settings into the QLineEdit(s) in the tab.

    Parameters:
    tab         The tab object.
    settings    The dictionary of settings.  This will be mutated.
    lines       A list of attribute names (str) of QLineEdits in the tab.
    """

    # set line edits
    for entry in lines:
        lineedit = getattr(tab, entry)
        settings[entry] = lineedit.text()


def load_lineedits(tab, settings, lines):
    """
    Load settings into the QLineEdit(s) in the tab.

    Parameters:
    tab         The tab object.
    settings    The dictionary of settings.
    lines       A list of attribute names (str) of QLineEdits in the tab.
    """

    log = logging.getLogger(__name__)

    # set line edits
    for entry in lines:
        lineedit = getattr(tab, entry)

        try:
            lineedit.setText(str(settings[entry]))
        except KeyError:
            log.warning("{0} criteria missing from settings file.".format(entry))


def save_textedits(tab, settings, edits):
    """
    Save settings into the QTextEdit(s) in the tab.

    Parameters:
    tab         The tab object.
    settings    The dictionary of settings.  This will be mutated.
    edits       A list of attribute names (str) of QTextEdits in the tab.
    """

    # set line edits
    for entry in edits:
        textedit = getattr(tab, entry)
        settings[entry] = textedit.toPlainText()


def load_textedits(tab, settings, edits):
    """
    Load settings into the QTextEdit(s) in the tab.

    Parameters:
    tab         The tab object.
    settings    The dictionary of settings.
    edits       A list of attribute names (str) of QTextEdits in the tab.
    """

    log = logging.getLogger(__name__)

    # set line edits
    for entry in edits:
        textedit = getattr(tab, entry)

        try:
            textedit.setPlainText(str(settings[entry]))
        except KeyError:
            log.warning("{0} criteria missing from settings file.".format(entry))


def save_listviews(tab, settings, listviews):
    """
    Save settings from the QListView selection(s) in the tab.

    Parameters:
    tab         The tab object.
    settings    The dictionary of settings.  This will be mutated.
    listviews   A list of attribute names (str) of QListViews in the tab.
    """

    for entry in listviews:
        listview = getattr(tab, entry)
        datamodel = listview.model()

        selections = []
        for index in listview.selectedIndexes():
            item = datamodel.data(index, Qt.DisplayRole)
            selections.append(item)

        settings[entry] = selections


def load_listviews(tab, settings, listviews):
    """
    Load settings into the QListView selection(s) in the tab.

    Parameters:
    tab         The tab object.
    settings    The dictionary of settings.
    listviews   A list of attribute names (str) of QListViews in the tab.
    """

    log = logging.getLogger(__name__)

    # set list selections
    for entry in listviews:
        try:
            selections = list(settings[entry])
        except KeyError:
            log.warning("{0} criteria missing from settings file.".format(entry))
            continue
        except TypeError:
            msg = "Invalid list for {0} criteria: {1!r}".format(entry, settings[entry])
            log.critical(msg)
            raise TypeError(msg)

        if not selections:
            continue

        listview = getattr(tab, entry)
        selectionmodel = listview.selectionModel()
        selectionmodel.clear()
        datamodel = listview.selectionModel().model()

        for row in range(datamodel.rowCount()):
            index = datamodel.createIndex(row, 0)
            item = datamodel.data(index, Qt.DisplayRole)

            if item in selections:
                selectionmodel.select(index, QItemSelectionModel.Select)


def save_comboboxes(tab, settings, comboboxes):
    """
    Save settings from the QComboBox selection in the tab.

    Parameters:
    tab         The tab object.
    settings    The dictionary of settings.  This will be mutated.
    comboboxes  A list of attribute names (str) of QListViews in the tab.
    """

    for entry in comboboxes:
        combobox = getattr(tab, entry)
        settings[entry] = combobox.currentData(Qt.DisplayRole)


def load_comboboxes(tab, settings, comboboxes):
    """
    Load settings into the QComboBox selection in the tab.

    Parameters:
    tab         The tab object.
    settings    The dictionary of settings.
    comboboxes  A list of attribute names (str) of QListViews in the tab.
    """

    log = logging.getLogger(__name__)

    for entry in comboboxes:
        try:
            selection = str(settings[entry])
        except KeyError:
            log.warning("{0} criteria missing from settings file.".format(entry))
            continue

        combobox = getattr(tab, entry)
        combobox.setCurrentText(selection)


def save_spinboxes(tab, settings, spinboxes):
    """
    Save settings from the QSpinBox value in the tab.

    Parameters:
    tab         The tab object.
    settings    The dictionary of settings.  This will be mutated.
    spinboxes   A list of attribute names (str) of QListViews in the tab.
    """

    for entry in spinboxes:
        spinbox = getattr(tab, entry)
        settings[entry] = spinbox.value()


def load_spinboxes(tab, settings, spinboxes):
    """
    Load settings into the QSpinBox value in the tab.

    Parameters:
    tab         The tab object.
    settings    The dictionary of settings.
    spinboxes   A list of attribute names (str) of QListViews in the tab.
    """

    log = logging.getLogger(__name__)

    for entry in spinboxes:
        try:
            value = int(settings[entry])
        except KeyError:
            log.warning("{0} criteria missing from settings file.".format(entry))
            continue
        except (ValueError, TypeError) as ex:
            msg = "Invalid value for {0} criteria: {1!r}.".format(entry, settings[entry])
            log.critical(msg)
            raise ex.__class__(msg)

        spinbox = getattr(tab, entry)
        spinbox.setValue(value)



# Copyright 2016, Tresys Technology, LLC
#
# This file is part of SETools.
#
# SETools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1 of
# the License, or (at your option) any later version.
#
# SETools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with SETools.  If not, see
# <http://www.gnu.org/licenses/>.
#
from PyQt5.QtCore import Qt, QModelIndex
from setools.exception import MLSDisabled

from .details import DetailsPopup
from .models import SEToolsTableModel


def user_detail(parent, user):
    """
    Create a dialog box for user details.

    Parameters:
    parent      The parent Qt Widget
    user        The user
    """

    detail = DetailsPopup(parent, "User detail: {0}".format(user))

    roles = sorted(user.roles)
    detail.append_header("Roles ({0}):".format(len(roles)))

    for role in roles:
        detail.append("    {0}".format(role))

    try:
        level = user.mls_level
        range_ = user.mls_range
    except MLSDisabled:
        pass
    else:
        detail.append_header("\nDefault MLS Level:")
        detail.append("    {0}".format(level))
        detail.append_header("\nMLS Range:")
        detail.append("    {0}".format(range_))

    detail.show()


class UserTableModel(SEToolsTableModel):

    """Table-based model for users."""

    headers = ["Name", "Roles", "Default Level", "Range"]

    def __init__(self, parent, mls):
        super(UserTableModel, self).__init__(parent)
        self.col_count = 4 if mls else 2

    def columnCount(self, parent=QModelIndex()):
        return self.col_count

    def data(self, index, role):
        if self.resultlist and index.isValid():
            if role == Qt.DisplayRole:
                row = index.row()
                col = index.column()
                user = self.resultlist[row]

                if col == 0:
                    return user.name
                elif col == 1:
                    return ", ".join(sorted(r.name for r in user.roles))
                elif col == 2:
                    try:
                        return str(user.mls_level)
                    except MLSDisabled:
                        return None
                elif col == 3:
                    try:
                        return str(user.mls_range)
                    except MLSDisabled:
                        return None

            elif role == Qt.UserRole:
                return user



# Copyright 2016, Tresys Technology, LLC
# Copyright 2018, Chris PeBenito <pebenito@ieee.org>
#
# This file is part of SETools.
#
# SETools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1 of
# the License, or (at your option) any later version.
#
# SETools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with SETools.  If not, see
# <http://www.gnu.org/licenses/>.
#
from collections import defaultdict
from typing import NamedTuple, Set

from ..policyrep import Category, Level, LevelDecl, Range, Sensitivity

from .descriptors import DiffResultDescriptor
from .difference import Difference, SymbolWrapper, Wrapper
from .typing import SymbolCache

_cats_cache: SymbolCache[Category] = defaultdict(dict)
_sens_cache: SymbolCache[Sensitivity] = defaultdict(dict)


class ModifiedCategory(NamedTuple):

    """Difference details for a modified category."""

    added_aliases: Set[str]
    removed_aliases: Set[str]
    matched_aliases: Set[str]


class ModifiedSensitivity(NamedTuple):

    """Difference details for a modified sensitivity."""

    added_aliases: Set[str]
    removed_aliases: Set[str]
    matched_aliases: Set[str]


class ModifiedLevelDecl(NamedTuple):

    """Difference details for a modified level declaration."""

    level: LevelDecl
    added_categories: Set[Category]
    removed_categories: Set[Category]
    matched_categories: Set[Category]


def category_wrapper_factory(category: Category) -> SymbolWrapper[Category]:
    """
    Wrap category from the specified policy.

    This caches results to prevent duplicate wrapper
    objects in memory.
    """
    try:
        return _cats_cache[category.policy][category]
    except KeyError:
        c = SymbolWrapper(category)
        _cats_cache[category.policy][category] = c
        return c


def sensitivity_wrapper_factory(sensitivity: Sensitivity) -> SymbolWrapper[Sensitivity]:
    """
    Wrap sensitivity from the specified policy.

    This caches results to prevent duplicate wrapper
    objects in memory.
    """
    try:
        return _sens_cache[sensitivity.policy][sensitivity]
    except KeyError:
        c = SymbolWrapper(sensitivity)
        _sens_cache[sensitivity.policy][sensitivity] = c
        return c


class CategoriesDifference(Difference):

    """Determine the difference in categories between two policies."""

    added_categories = DiffResultDescriptor("diff_categories")
    removed_categories = DiffResultDescriptor("diff_categories")
    modified_categories = DiffResultDescriptor("diff_categories")

    def diff_categories(self) -> None:
        """Generate the difference in categories between the policies."""

        self.log.info(
            "Generating category differences from {0.left_policy} to {0.right_policy}".format(self))

        self.added_categories, self.removed_categories, matched_categories = self._set_diff(
            (category_wrapper_factory(c) for c in self.left_policy.categories()),
            (category_wrapper_factory(c) for c in self.right_policy.categories()))

        self.modified_categories = dict()

        for left_category, right_category in matched_categories:
            # Criteria for modified categories
            # 1. change to aliases
            added_aliases, removed_aliases, matched_aliases = self._set_diff(
                left_category.aliases(), right_category.aliases(), unwrap=False)

            if added_aliases or removed_aliases:
                self.modified_categories[left_category] = ModifiedCategory(added_aliases,
                                                                           removed_aliases,
                                                                           matched_aliases)

    #
    # Internal functions
    #
    def _reset_diff(self) -> None:
        """Reset diff results on policy changes."""
        self.log.debug("Resetting category differences")
        self.added_categories = None
        self.removed_categories = None
        self.modified_categories = None


class SensitivitiesDifference(Difference):

    """Determine the difference in sensitivities between two policies."""

    added_sensitivities = DiffResultDescriptor("diff_sensitivities")
    removed_sensitivities = DiffResultDescriptor("diff_sensitivities")
    modified_sensitivities = DiffResultDescriptor("diff_sensitivities")

    def diff_sensitivities(self) -> None:
        """Generate the difference in sensitivities between the policies."""

        self.log.info(
            "Generating sensitivity differences from {0.left_policy} to {0.right_policy}".
            format(self))

        self.added_sensitivities, self.removed_sensitivities, matched_sensitivities = \
            self._set_diff(
                (sensitivity_wrapper_factory(s) for s in self.left_policy.sensitivities()),
                (sensitivity_wrapper_factory(s) for s in self.right_policy.sensitivities()))

        self.modified_sensitivities = dict()

        for left_sens, right_sens in matched_sensitivities:
            # Criteria for modified sensitivities
            # 1. change to aliases
            added_aliases, removed_aliases, matched_aliases = self._set_diff(
                left_sens.aliases(), right_sens.aliases(), unwrap=False)

            if added_aliases or removed_aliases:
                self.modified_sensitivities[left_sens] = ModifiedSensitivity(added_aliases,
                                                                             removed_aliases,
                                                                             matched_aliases)

    #
    # Internal functions
    #
    def _reset_diff(self) -> None:
        """Reset diff results on policy changes."""
        self.log.debug("Resetting sensitivity differences")
        self.added_sensitivities = None
        self.removed_sensitivities = None
        self.modified_sensitivities = None


class LevelDeclsDifference(Difference):

    """Determine the difference in levels between two policies."""

    added_levels = DiffResultDescriptor("diff_levels")
    removed_levels = DiffResultDescriptor("diff_levels")
    modified_levels = DiffResultDescriptor("diff_levels")

    def diff_levels(self) -> None:
        """Generate the difference in levels between the policies."""

        self.log.info(
            "Generating level decl differences from {0.left_policy} to {0.right_policy}".
            format(self))

        self.added_levels, self.removed_levels, matched_levels = \
            self._set_diff(
                (LevelDeclWrapper(s) for s in self.left_policy.levels()),
                (LevelDeclWrapper(s) for s in self.right_policy.levels()))

        self.modified_levels = []

        for left_level, right_level in matched_levels:
            # Criteria for modified levels
            # 1. change to allowed categories
            added_categories, removed_categories, matched_categories = self._set_diff(
                (category_wrapper_factory(c) for c in left_level.categories()),
                (category_wrapper_factory(c) for c in right_level.categories()))

            if added_categories or removed_categories:
                self.modified_levels.append(ModifiedLevelDecl(
                    left_level, added_categories, removed_categories, matched_categories))

    #
    # Internal functions
    #
    def _reset_diff(self) -> None:
        """Reset diff results on policy changes."""
        self.log.debug("Resetting sensitivity differences")
        self.added_levels = None
        self.removed_levels = None
        self.modified_levels = None


# Pylint bug: https://github.com/PyCQA/pylint/issues/2822
class LevelDeclWrapper(Wrapper[LevelDecl]):  # pylint: disable=unsubscriptable-object

    """Wrap level declarations to allow comparisons."""

    __slots__ = ("sensitivity",)

    def __init__(self, level: LevelDecl) -> None:
        self.origin = level
        self.sensitivity = sensitivity_wrapper_factory(level.sensitivity)
        self.key = hash(level)

    def __hash__(self):
        return self.key

    def __eq__(self, other):
        # non-MLS policies have no level declarations so there
        # should be no AttributeError possiblity here
        return self.sensitivity == other.sensitivity

    def __lt__(self, other):
        return self.sensitivity < other.sensitivity


# Pylint bug: https://github.com/PyCQA/pylint/issues/2822
class LevelWrapper(Wrapper[Level]):  # pylint: disable=unsubscriptable-object

    """Wrap levels to allow comparisons."""

    __slots__ = ("sensitivity", "categories")

    def __init__(self, level: Level) -> None:
        self.origin = level
        self.sensitivity = sensitivity_wrapper_factory(level.sensitivity)
        self.categories = set(category_wrapper_factory(c) for c in level.categories())

    def __hash__(self):
        return hash(self.origin)

    def __eq__(self, other):
        try:
            return self.sensitivity == other.sensitivity and \
                self.categories == other.categories
        except AttributeError:
            # comparing an MLS policy to non-MLS policy will result in
            # other being None
            return False

    def __lt__(self, other):
        try:
            return self.sensitivity < other.sensitivity and \
                self.categories < other.categories
        except AttributeError:
            # comparing an MLS policy to non-MLS policy will result in
            # other being None
            return False


# Pylint bug: https://github.com/PyCQA/pylint/issues/2822
class RangeWrapper(Wrapper[Range]):  # pylint: disable=unsubscriptable-object

    """
    Wrap ranges to allow comparisons.

    This only compares the low and high levels of the range.
    It does not detect additions/removals/modifications
    to levels between the low and high levels of the range.
    """

    __slots__ = ("low", "high")

    def __init__(self, range_: Range) -> None:
        self.origin = range_
        self.low = LevelWrapper(range_.low)
        self.high = LevelWrapper(range_.high)

    def __hash__(self):
        return hash(self.origin)

    def __eq__(self, other):
        try:
            return self.low == other.low and \
                self.high == other.high
        except AttributeError:
            # comparing an MLS policy to non-MLS policy will result in
            # other being None
            return False

    def __lt__(self, other):
        try:
            return self.low < other.low and \
                self.high < other.high
        except AttributeError:
            # comparing an MLS policy to non-MLS policy will result in
            # other being None
            return False



# Copyright 2016, Tresys Technology, LLC
# Copyright 2018, Chris PeBenito <pebenito@ieee.org>
#
# This file is part of SETools.
#
# SETools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1 of
# the License, or (at your option) any later version.
#
# SETools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with SETools.  If not, see
# <http://www.gnu.org/licenses/>.
#

import logging

from PyQt5.QtCore import Qt, QSortFilterProxyModel, QStringListModel, QThread
from PyQt5.QtGui import QPalette, QTextCursor
from PyQt5.QtWidgets import QCompleter, QHeaderView, QMessageBox, QProgressDialog
from setools import MLSRuleQuery

from ..logtosignal import LogHandlerToSignal
from ..models import SEToolsListModel, invert_list_selection
from ..mlsrulemodel import MLSRuleTableModel
from .analysistab import AnalysisSection, AnalysisTab
from .queryupdater import QueryResultsUpdater
from .workspace import load_checkboxes, load_textedits, save_checkboxes, save_textedits


class SummaryTab(AnalysisTab):

    """An SELinux policy summary."""

    section = AnalysisSection.General
    tab_title = "Summary"
    mlsonly = False

    def __init__(self, parent, policy, perm_map):
        super(SummaryTab, self).__init__(parent)
        self.log = logging.getLogger(__name__)
        self.setupUi(policy)

    def setupUi(self, p):
        self.load_ui("apol/summary.ui")

        # Ensure settings are consistent with the initial .ui state
        self.notes.setHidden(not self.notes_expander.isChecked())

        # Populate policy stats
        mls = "enabled" if p.mls else "disabled"

        self.policy_version.setText(str(p.version))
        self.mls.setText(mls)
        self.handle_unknown.setText(p.handle_unknown.name)
        self.class_count.setText(str(p.class_count))
        self.perms_count.setText(str(p.permission_count))
        self.type_count.setText(str(p.type_count))
        self.attribute_count.setText(str(p.type_attribute_count))
        self.user_count.setText(str(p.user_count))
        self.role_count.setText(str(p.role_count))
        self.bool_count.setText(str(p.boolean_count))
        self.allow_count.setText(str(p.allow_count))
        self.neverallow_count.setText(str(p.neverallow_count))
        self.auditallow_count.setText(str(p.auditallow_count))
        self.dontaudit_count.setText(str(p.dontaudit_count))
        self.type_transition_count.setText(str(p.type_transition_count))
        self.type_change_count.setText(str(p.type_change_count))
        self.type_member_count.setText(str(p.type_member_count))
        self.role_allow_count.setText(str(p.role_allow_count))
        self.role_transition_count.setText(str(p.role_transition_count))
        self.constrain_count.setText(str(p.constraint_count))
        self.validatetrans_count.setText(str(p.validatetrans_count))
        self.permissive_count.setText(str(p.permissives_count))
        self.default_count.setText(str(p.default_count))
        self.typebounds_count.setText(str(p.typebounds_count))
        self.allowxperm_count.setText(str(p.allowxperm_count))
        self.neverallowxperm_count.setText(str(p.neverallowxperm_count))
        self.auditallowxperm_count.setText(str(p.auditallowxperm_count))
        self.dontauditxperm_count.setText(str(p.dontauditxperm_count))
        self.ibendportcon_count.setText(str(p.ibendportcon_count))
        self.ibpkeycon_count.setText(str(p.ibpkeycon_count))
        self.initsid_count.setText(str(p.initialsids_count))
        self.fs_use_count.setText(str(p.fs_use_count))
        self.genfscon_count.setText(str(p.genfscon_count))
        self.portcon_count.setText(str(p.portcon_count))
        self.netifcon_count.setText(str(p.netifcon_count))
        self.nodecon_count.setText(str(p.nodecon_count))

        if p.mls:
            self.sensitivity_count.setText(str(p.level_count))
            self.category_count.setText(str(p.category_count))
            self.range_transition_count.setText(str(p.range_transition_count))
            self.mlsconstrain_count.setText(str(p.mlsconstraint_count))
            self.mlsvalidatetrans_count.setText(str(p.mlsvalidatetrans_count))
        else:
            self.sens_label.setEnabled(False)
            self.sens_label.setToolTip("MLS is disabled in this policy.")
            self.cats_label.setEnabled(False)
            self.cats_label.setToolTip("MLS is disabled in this policy.")
            self.range_trans_label.setEnabled(False)
            self.range_trans_label.setToolTip("MLS is disabled in this policy.")
            self.mlsconstrain_label.setEnabled(False)
            self.mlsconstrain_label.setToolTip("MLS is disabled in this policy.")
            self.mlsvalidatetrans_label.setEnabled(False)
            self.mlsvalidatetrans_label.setToolTip("MLS is disabled in this policy.")
            self.sensitivity_count.setEnabled(False)
            self.sensitivity_count.setToolTip("MLS is disabled in this policy.")
            self.category_count.setEnabled(False)
            self.category_count.setToolTip("MLS is disabled in this policy.")
            self.range_transition_count.setEnabled(False)
            self.range_transition_count.setToolTip("MLS is disabled in this policy.")
            self.mlsconstrain_count.setEnabled(False)
            self.mlsconstrain_count.setToolTip("MLS is disabled in this policy.")
            self.mlsvalidatetrans_count.setEnabled(False)
            self.mlsvalidatetrans_count.setToolTip("MLS is disabled in this policy.")

        # Fill policy capabilities list
        self.polcaps.addItems([str(c) for c in p.polcaps()])

    #
    # Save/Load tab
    #
    def save(self):
        """Return a dictionary of settings."""
        settings = {}
        save_checkboxes(self, settings, ["notes_expander"])
        save_textedits(self, settings, ["notes"])
        return settings

    def load(self, settings):
        load_checkboxes(self, settings, ["notes_expander"])
        load_textedits(self, settings, ["notes"])



# Copyright 2015, Tresys Technology, LLC
# Copyright 2018, Chris PeBenito <pebenito@ieee.org>
#
# This file is part of SETools.
#
# SETools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1 of
# the License, or (at your option) any later version.
#
# SETools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with SETools.  If not, see
# <http://www.gnu.org/licenses/>.
#
from collections import defaultdict
from typing import NamedTuple, Set, Union

from ..policyrep import Type, TypeAttribute, TypeOrAttr

from .descriptors import DiffResultDescriptor
from .difference import Difference, SymbolWrapper
from .typeattr import typeattr_wrapper_factory
from .typing import SymbolCache

_types_cache: SymbolCache[Type] = defaultdict(dict)


class ModifiedType(NamedTuple):

    """Difference details for a modified type."""

    added_attributes: Set[TypeAttribute]
    removed_attributes: Set[TypeAttribute]
    matched_attributes: Set[TypeAttribute]
    modified_permissive: bool
    permissive: bool
    added_aliases: Set[str]
    removed_aliases: Set[str]
    matched_aliases: Set[str]


def type_wrapper_factory(type_: Type) -> SymbolWrapper[Type]:
    """
    Wrap types from the specified policy.

    This caches results to prevent duplicate wrapper
    objects in memory.
    """
    try:
        return _types_cache[type_.policy][type_]
    except KeyError:
        t = SymbolWrapper(type_)
        _types_cache[type_.policy][type_] = t
        return t


def type_or_attr_wrapper_factory(type_: TypeOrAttr) -> \
        Union[SymbolWrapper[Type], SymbolWrapper[TypeAttribute]]:

    """
    Wrap types or attributes from the specified policy.

    This caches results to prevent duplicate wrapper
    objects in memory.
    """
    if isinstance(type_, Type):
        return type_wrapper_factory(type_)
    else:
        return typeattr_wrapper_factory(type_)


class TypesDifference(Difference):

    """Determine the difference in types between two policies."""

    added_types = DiffResultDescriptor("diff_types")
    removed_types = DiffResultDescriptor("diff_types")
    modified_types = DiffResultDescriptor("diff_types")

    def diff_types(self) -> None:
        """Generate the difference in types between the policies."""

        self.log.info(
            "Generating type differences from {0.left_policy} to {0.right_policy}".format(self))

        self.added_types, self.removed_types, matched_types = self._set_diff(
            (SymbolWrapper(t) for t in self.left_policy.types()),
            (SymbolWrapper(t) for t in self.right_policy.types()))

        self.modified_types = dict()

        for left_type, right_type in matched_types:
            # Criteria for modified types
            # 1. change to attribute set, or
            # 2. change to alias set, or
            # 3. different permissive setting
            added_attr, removed_attr, matched_attr = self._set_diff(
                (SymbolWrapper(a) for a in left_type.attributes()),
                (SymbolWrapper(a) for a in right_type.attributes()))

            added_aliases, removed_aliases, matched_aliases = self._set_diff(left_type.aliases(),
                                                                             right_type.aliases(),
                                                                             unwrap=False)

            left_permissive = left_type.ispermissive
            right_permissive = right_type.ispermissive
            mod_permissive = left_permissive != right_permissive

            if added_attr or removed_attr or added_aliases or removed_aliases or mod_permissive:
                self.modified_types[left_type] = ModifiedType(added_attr,
                                                              removed_attr,
                                                              matched_attr,
                                                              mod_permissive,
                                                              left_permissive,
                                                              added_aliases,
                                                              removed_aliases,
                                                              matched_aliases)

    #
    # Internal functions
    #
    def _reset_diff(self) -> None:
        """Reset diff results on policy changes."""
        self.log.debug("Resetting type differences")
        self.added_types = None
        self.removed_types = None
        self.modified_types = None



# Copyright 2016, Tresys Technology, LLC
#
# This file is part of SETools.
#
# SETools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1 of
# the License, or (at your option) any later version.
#
# SETools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with SETools.  If not, see
# <http://www.gnu.org/licenses/>.
#
import logging
import copy

from PyQt5.QtCore import pyqtSignal, pyqtSlot, Qt
from PyQt5.QtGui import QPalette
from PyQt5.QtWidgets import QDialog, QFrame, QWidget

from ..models import SEToolsListModel
from ..widget import SEToolsWidget


class PermissionMapEditor(SEToolsWidget, QDialog):

    """
    A permission map editor.  This dialog has two versions,
    one for editing the weight/direction and another for
    including or excluding permissions in an analysis.

    Parameters:
    parent      The parent Qt widget
    edit        (bool) If true, the dialog will take
                the editor behavior.  If False, the dialog
                will take the enable/disable permission
                behavior.
    """

    class_toggle = pyqtSignal(bool)

    def __init__(self, parent, edit):
        super(PermissionMapEditor, self).__init__(parent)
        self.log = logging.getLogger(__name__)
        self.parent = parent
        self.edit = edit
        self.setupUi()

    def setupUi(self):
        self.load_ui("apol/permmap_editor.ui")

        # set up class list
        self.class_model = SEToolsListModel(self)
        self.classes.setModel(self.class_model)

        # permission widgets
        self.widgets = []

        # set up editor mode
        self.enable_all.setHidden(self.edit)
        self.disable_all.setHidden(self.edit)

        # connect signals
        self.classes.selectionModel().selectionChanged.connect(self.class_selected)
        self.enable_all.clicked.connect(self.enable_all_perms)
        self.disable_all.clicked.connect(self.disable_all_perms)

    def show(self, perm_map):
        # keep an internal copy because the map is mutable
        # and this dialog may be canceled after some edits.
        self.perm_map = copy.deepcopy(perm_map)

        self.class_model.item_list = sorted(perm_map.classes())

        # clear class selection and mappings
        # since this widget will typically
        # be reused.
        self.classes.clearSelection()
        self._clear_mappings()
        self.enable_all.setToolTip(None)
        self.disable_all.setToolTip(None)

        if self.edit:
            self.setWindowTitle("{0} - Permission Map Editor - apol".format(self.perm_map))
        else:
            self.setWindowTitle("{0} - Permission Map Viewer - apol".format(self.perm_map))

        super(PermissionMapEditor, self).show()

    def accept(self):
        self.parent.apply_permmap(self.perm_map)
        super(PermissionMapEditor, self).accept()

    def class_selected(self):
        # the .ui is set to 1 selection
        for index in self.classes.selectionModel().selectedIndexes():
            class_name = self.class_model.data(index, Qt.DisplayRole)

        self.log.debug("Setting class to {0}".format(class_name))

        self.enable_all.setToolTip("Include all permissions in the {0} class.".format(class_name))
        self.disable_all.setToolTip("Exclude all permissions in the {0} class.".format(class_name))

        self._clear_mappings()

        # populate new mappings
        for perm in sorted(self.perm_map.perms(class_name)):
            # create permission mapping
            mapping = PermissionMapping(self, perm, self.edit)
            mapping.setAttribute(Qt.WA_DeleteOnClose)
            self.class_toggle.connect(mapping.enabled.setChecked)
            self.perm_mappings.addWidget(mapping)
            self.widgets.append(mapping)

            # add horizonal line
            line = QFrame(self)
            line.setFrameShape(QFrame.HLine)
            line.setFrameShadow(QFrame.Sunken)
            self.perm_mappings.addWidget(line)
            self.widgets.append(line)

    def enable_all_perms(self):
        self.class_toggle.emit(True)

    def disable_all_perms(self):
        self.class_toggle.emit(False)

    #
    # Internal functions
    #
    def _clear_mappings(self):
        # delete current mappings
        for mapping in self.widgets:
            mapping.close()

        self.widgets.clear()


index_to_setting = ["r", "w", "b", "n"]
index_to_word = ["Read", "Write", "Both", "None"]
setting_to_index = {"r": 0, "w": 1, "b": 2, "n": 3}


class PermissionMapping(SEToolsWidget, QWidget):

    """
    A widget representing mapping for a particular permission.
    This dialog has two versions, one for editing the weight/direction
    and another for including or excluding permissions in an analysis.

    Parameters:
    parent      The parent Qt widget
    edit        (bool) If true, the widget will take
                the editor behavior.  If False, the dialog
                will take the enable/disable permission
                behavior.
    """

    def __init__(self, parent, mapping, edit):
        super(PermissionMapping, self).__init__(parent)
        self.log = logging.getLogger(__name__)
        self.parent = parent
        self.mapping = mapping
        self.edit = edit
        self.setupUi()

    def setupUi(self):
        self.load_ui("apol/permmapping.ui")

        self.permission.setText(str(self.mapping.perm))
        self.weight.setValue(self.mapping.weight)
        self.enabled.setChecked(self.mapping.enabled)

        if self.edit:
            self.weight.setToolTip("Set the information flow weight of {0}".format(
                self.mapping.perm))
            self.direction.setToolTip("Set the information flow direction of {0}".format(
                self.mapping.perm))
        else:
            self.enabled.setToolTip("Include or exclude {0} from the analysis.".format(
                self.mapping.perm))

        self.weight.setEnabled(self.edit)
        self.direction.setEnabled(self.edit)
        self.enabled.setHidden(self.edit)

        # setup color palettes for direction
        self.orig_palette = self.direction.palette()
        self.error_palette = self.direction.palette()
        self.error_palette.setColor(QPalette.Button, Qt.red)
        self.error_palette.setColor(QPalette.ButtonText, Qt.white)

        # setup direction
        self.direction.insertItems(0, index_to_word)
        if self.mapping.direction == 'u':
            # Temporarily add unmapped value to items
            self.direction.insertItem(len(index_to_word), "Unmapped")
            self.direction.setCurrentText("Unmapped")
            self.direction.setPalette(self.error_palette)
            self.unmapped = True
        else:
            self.direction.setCurrentIndex(setting_to_index[self.mapping.direction])
            self.unmapped = False

        # connect signals
        self.direction.currentIndexChanged.connect(self.set_direction)
        self.weight.valueChanged.connect(self.set_weight)
        self.enabled.toggled.connect(self.set_enabled)

    def set_direction(self, value):
        if self.unmapped:
            if value == "Unmapped":
                return

            # Remove unmapped item if setting the mapping.
            self.direction.removeItem(len(index_to_word))
            self.direction.setPalette(self.orig_palette)
            self.unmapped = False

        dir_ = index_to_setting[value]
        self.log.debug("Setting {0.class_}:{0.perm} direction to {1}".format(self.mapping, dir_))
        self.mapping.direction = dir_

    def set_weight(self, value):
        self.log.debug("Setting {0.class_}:{0.perm} weight to {1}".format(self.mapping, value))
        self.mapping.weight = int(value)

    def set_enabled(self, value):
        self.log.debug("Setting {0.class_}:{0.perm} enabled to {1}".format(self.mapping, value))
        self.mapping.enabled = value



# Copyright 2015, Tresys Technology, LLC
# Copyright 2018, Chris PeBenito <pebenito@ieee.org>
#
# This file is part of SETools.
#
# SETools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1 of
# the License, or (at your option) any later version.
#
# SETools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with SETools.  If not, see
# <http://www.gnu.org/licenses/>.
#
from collections import defaultdict
from contextlib import suppress
from typing import NamedTuple, Set

from ..exception import NoCommon
from ..policyrep import ObjClass

from .descriptors import DiffResultDescriptor
from .difference import Difference, SymbolWrapper
from .typing import SymbolCache

_class_cache: SymbolCache[ObjClass] = defaultdict(dict)


class ModifiedObjClass(NamedTuple):

    """Difference details for a modified object class."""

    added_perms: Set[str]
    removed_perms: Set[str]
    matched_perms: Set[str]


def class_wrapper_factory(class_: ObjClass) -> SymbolWrapper[ObjClass]:
    """
    Wrap class from the specified policy.

    This caches results to prevent duplicate wrapper
    objects in memory.
    """

    try:
        return _class_cache[class_.policy][class_]
    except KeyError:
        c = SymbolWrapper(class_)
        _class_cache[class_.policy][class_] = c
        return c


class ObjClassDifference(Difference):

    """
    Determine the difference in object classes
    between two policies.
    """

    added_classes = DiffResultDescriptor("diff_classes")
    removed_classes = DiffResultDescriptor("diff_classes")
    modified_classes = DiffResultDescriptor("diff_classes")

    def diff_classes(self) -> None:
        """Generate the difference in object classes between the policies."""

        self.log.info(
            "Generating class differences from {0.left_policy} to {0.right_policy}".format(self))

        self.added_classes, self.removed_classes, matched_classes = self._set_diff(
            (SymbolWrapper(c) for c in self.left_policy.classes()),
            (SymbolWrapper(c) for c in self.right_policy.classes()))

        self.modified_classes = dict()

        for left_class, right_class in matched_classes:
            # Criteria for modified classes
            # 1. change to permissions (inherited common is expanded)

            left_perms = left_class.perms
            with suppress(NoCommon):
                left_perms |= left_class.common.perms

            right_perms = right_class.perms
            with suppress(NoCommon):
                right_perms |= right_class.common.perms

            added_perms, removed_perms, matched_perms = self._set_diff(left_perms,
                                                                       right_perms,
                                                                       unwrap=False)

            if added_perms or removed_perms:
                self.modified_classes[left_class] = ModifiedObjClass(added_perms,
                                                                     removed_perms,
                                                                     matched_perms)

    #
    # Internal functions
    #
    def _reset_diff(self) -> None:
        """Reset diff results on policy changes."""
        self.log.debug("Resetting object class differences")
        self.added_classes = None
        self.removed_classes = None
        self.modified_classes = None



# Copyright 2016, Tresys Technology, LLC
#
# This file is part of SETools.
#
# SETools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1 of
# the License, or (at your option) any later version.
#
# SETools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with SETools.  If not, see
# <http://www.gnu.org/licenses/>.
#
from typing import NamedTuple

from ..policyrep import Context

from .context import ContextWrapper
from .descriptors import DiffResultDescriptor
from .difference import Difference, SymbolWrapper


class ModifiedInitialSID(NamedTuple):

    """Difference details for a modified initial SID."""

    added_context: Context
    removed_context: Context


class InitialSIDsDifference(Difference):

    """Determine the difference in initsids between two policies."""

    added_initialsids = DiffResultDescriptor("diff_initialsids")
    removed_initialsids = DiffResultDescriptor("diff_initialsids")
    modified_initialsids = DiffResultDescriptor("diff_initialsids")

    def diff_initialsids(self) -> None:
        """Generate the difference in initial SIDs between the policies."""

        self.log.info("Generating initial SID differences from {0.left_policy} to {0.right_policy}".
                      format(self))

        self.added_initialsids, self.removed_initialsids, matched_initialsids = self._set_diff(
            (SymbolWrapper(i) for i in self.left_policy.initialsids()),
            (SymbolWrapper(i) for i in self.right_policy.initialsids()))

        self.modified_initialsids = dict()

        for left_initialsid, right_initialsid in matched_initialsids:
            # Criteria for modified initialsids
            # 1. change to context
            if ContextWrapper(left_initialsid.context) != ContextWrapper(right_initialsid.context):
                self.modified_initialsids[left_initialsid] = ModifiedInitialSID(
                    right_initialsid.context, left_initialsid.context)

    #
    # Internal functions
    #
    def _reset_diff(self) -> None:
        """Reset diff results on policy changes."""
        self.log.debug("Resetting initialsid differences")
        self.added_initialsids = None
        self.removed_initialsids = None
        self.modified_initialsids = None



# Copyright 2015-2016, Tresys Technology, LLC
# Copyright 2016, 2019, Chris PeBenito <pebenito@ieee.org>
#
# This file is part of SETools.
#
# SETools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1 of
# the License, or (at your option) any later version.
#
# SETools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with SETools.  If not, see
# <http://www.gnu.org/licenses/>.
#
import os
import shutil
import logging
import json
from contextlib import suppress

import pkg_resources
from PyQt5.QtCore import pyqtSlot, Qt, QProcess
from PyQt5.QtWidgets import QApplication, QFileDialog, QLineEdit, QMainWindow, QMessageBox
from setools import __version__, PermissionMap, SELinuxPolicy

from ..widget import SEToolsWidget
from ..logtosignal import LogHandlerToSignal
from .analysistab import TAB_REGISTRY
from .chooseanalysis import ChooseAnalysis
from .config import ApolConfig
from .exception import TabFieldError
from .permmapedit import PermissionMapEditor
from .summary import SummaryTab


BIN_SEARCH_PATHS = ("/usr/local/bin:/usr/bin:/bin")
POSSIBLE_ASSISTANT = ("assistant", "assistant-qt5")


class ApolMainWindow(SEToolsWidget, QMainWindow):

    def __init__(self, filename):
        super(ApolMainWindow, self).__init__()
        self.log = logging.getLogger(__name__)
        self._permmap = None
        self._policy = None
        self.config = ApolConfig()
        self.setupUi()

        self.load_permmap()

        if filename:
            self.load_policy(filename)

        if self._policy:
            self.create_new_analysis("Summary", SummaryTab)

        self.update_window_title()
        self.toggle_workspace_actions()

    def setupUi(self):
        self.load_ui("apol/apol.ui")

        self.tab_counter = 0

        # set up analysis menu
        self.chooser = ChooseAnalysis(self)

        # set up error message dialog
        self.error_msg = QMessageBox(self)
        self.error_msg.setStandardButtons(QMessageBox.Ok)

        # set up permission map editor
        self.permmap_editor = PermissionMapEditor(self, True)

        # set up tab name editor
        self.tab_editor = QLineEdit(self.AnalysisTabs)
        self.tab_editor.setWindowFlags(Qt.Popup)

        # configure tab bar context menu
        tabBar = self.AnalysisTabs.tabBar()
        tabBar.addAction(self.rename_tab_action)
        tabBar.addAction(self.close_tab_action)
        tabBar.setContextMenuPolicy(Qt.ActionsContextMenu)

        # capture INFO and higher Python messages from setools lib for status bar
        handler = LogHandlerToSignal()
        handler.message.connect(self.statusbar.showMessage)
        logging.getLogger("setools").addHandler(handler)
        logging.getLogger("setoolsgui").addHandler(handler)

        # set up help browser process
        self.help_process = QProcess()

        # connect signals
        self.open_policy.triggered.connect(self.select_policy)
        self.close_policy_action.triggered.connect(self.close_policy)
        self.open_permmap.triggered.connect(self.select_permmap)
        self.new_analysis.triggered.connect(self.choose_analysis)
        self.AnalysisTabs.currentChanged.connect(self.toggle_workspace_actions)
        self.AnalysisTabs.tabCloseRequested.connect(self.close_tab)
        self.AnalysisTabs.tabBarDoubleClicked.connect(self.tab_name_editor)
        self.tab_editor.editingFinished.connect(self.rename_tab)
        self.rename_tab_action.triggered.connect(self.rename_active_tab)
        self.close_tab_action.triggered.connect(self.close_active_tab)
        self.new_from_settings_action.triggered.connect(self.new_analysis_from_config)
        self.load_settings_action.triggered.connect(self.load_settings)
        self.save_settings_action.triggered.connect(self.save_settings)
        self.load_workspace_action.triggered.connect(self.load_workspace)
        self.save_workspace_action.triggered.connect(self.save_workspace)
        self.copy_action.triggered.connect(self.copy)
        self.cut_action.triggered.connect(self.cut)
        self.paste_action.triggered.connect(self.paste)
        self.edit_permmap_action.triggered.connect(self.edit_permmap)
        self.save_permmap_action.triggered.connect(self.save_permmap)
        self.about_apol_action.triggered.connect(self.about_apol)
        self.apol_help_action.triggered.connect(self.apol_help)
        self.help_process.errorOccurred.connect(self.help_failed)

        self.show()

    def update_window_title(self):
        if self._policy:
            self.setWindowTitle("{0} - apol".format(self._policy))
        else:
            self.setWindowTitle("apol")

    #
    # Policy handling
    #
    def select_policy(self):
        old_policy = self._policy

        if old_policy and self.AnalysisTabs.count() > 0:
            reply = QMessageBox.question(
                self, "Continue?",
                "Loading a policy will close all existing analyses.  Continue?",
                QMessageBox.Yes | QMessageBox.No)

            if reply == QMessageBox.No:
                return

        filename = QFileDialog.getOpenFileName(self, "Open policy file", ".",
                                               "SELinux Policies (policy.* sepolicy);;"
                                               "All Files (*)")[0]
        if filename:
            self.load_policy(filename)

        if self._policy != old_policy:
            # policy loading succeeded, clear any
            # existing tabs
            self.AnalysisTabs.clear()
            self.create_new_analysis("Summary", SummaryTab)

    def load_policy(self, filename):
        try:
            self._policy = SELinuxPolicy(filename)
        except Exception as ex:
            self.log.critical("Failed to load policy \"{0}\"".format(filename))
            self.error_msg.critical(self, "Policy loading error", str(ex))
        else:
            self.update_window_title()
            self.toggle_workspace_actions()

            if self._permmap:
                self._permmap.map_policy(self._policy)
                self.apply_permmap()

    def close_policy(self):
        if self.AnalysisTabs.count() > 0:
            reply = QMessageBox.question(
                self, "Continue?",
                "Closing a policy will close all existing analyses.  Continue?",
                QMessageBox.Yes | QMessageBox.No)

            if reply == QMessageBox.No:
                return

        self.AnalysisTabs.clear()
        self._policy = None
        self.update_window_title()
        self.toggle_workspace_actions()

    #
    # Permission map handling
    #
    def select_permmap(self):
        filename = QFileDialog.getOpenFileName(self, "Open permission map file", ".")[0]
        if filename:
            self.load_permmap(filename)

    def load_permmap(self, filename=None):
        try:
            self._permmap = PermissionMap(filename)
        except Exception as ex:
            self.log.critical("Failed to load default permission map: {0}".format(ex))
            self.error_msg.critical(self, "Permission map loading error", str(ex))
        else:
            if self._policy:
                self._permmap.map_policy(self._policy)
                self.apply_permmap()

    def edit_permmap(self):
        if not self._permmap:
            self.error_msg.critical(self, "No open permission map",
                                    "Cannot edit permission map. Please open a map first.")
            self.select_permmap()

        # in case user cancels out of
        # choosing a permmap, recheck
        if self._permmap:
            self.permmap_editor.show(self._permmap)

    def apply_permmap(self, perm_map=None):
        if perm_map:
            self._permmap = perm_map

        for index in range(self.AnalysisTabs.count()):
            tab = self.AnalysisTabs.widget(index)
            self.log.debug("Updating permmap in tab {0} ({1}: \"{2}\")".format(
                           index, tab, tab.objectName()))
            tab.perm_map = self._permmap

    def save_permmap(self):
        path = str(self._permmap) if self._permmap else "perm_map"
        filename = QFileDialog.getSaveFileName(self, "Save permission map file", path)[0]
        if filename:
            try:
                self._permmap.save(filename)
            except Exception as ex:
                self.log.critical("Failed to save permission map: {0}".format(ex))
                self.error_msg.critical(self, "Permission map saving error", str(ex))

    #
    # Analysis tab handling
    #
    def choose_analysis(self):
        if not self._policy:
            self.error_msg.critical(self, "No open policy",
                                    "Cannot start a new analysis. Please open a policy first.")

            self.select_policy()

        if self._policy:
            # this check of self._policy is here in case someone
            # tries to start an analysis with no policy open, but then
            # cancels out of the policy file chooser or there is an
            # error opening the policy file.
            self.chooser.show(self._policy.mls)

    def create_new_analysis(self, tabtitle, tabclass):
        self.tab_counter += 1
        counted_name = "{0}: {1}".format(self.tab_counter, tabtitle)

        newanalysis = tabclass(self, self._policy, self._permmap)
        newanalysis.setAttribute(Qt.WA_DeleteOnClose)
        newanalysis.setObjectName(counted_name)

        index = self.AnalysisTabs.addTab(newanalysis, counted_name)
        self.AnalysisTabs.setTabToolTip(index, tabtitle)
        self.AnalysisTabs.setCurrentIndex(index)
        return index

    def tab_name_editor(self, index):
        if index >= 0:
            tab_area = self.AnalysisTabs.tabBar().tabRect(index)
            self.tab_editor.move(self.AnalysisTabs.mapToGlobal(tab_area.topLeft()))
            self.tab_editor.setText(self.AnalysisTabs.tabText(index))
            self.tab_editor.selectAll()
            self.tab_editor.show()
            self.tab_editor.setFocus()

    def close_active_tab(self):
        """Close the active tab. This is called from the context menu."""
        index = self.AnalysisTabs.currentIndex()
        if index >= 0:
            self.close_tab(index)

    def rename_active_tab(self):
        """Rename the active tab."""
        index = self.AnalysisTabs.currentIndex()
        if index >= 0:
            self.tab_name_editor(index)

    def close_tab(self, index):
        """Close a tab specified by index."""
        widget = self.AnalysisTabs.widget(index)
        widget.close()
        self.AnalysisTabs.removeTab(index)

    def rename_tab(self):
        # this should never be negative since the editor is modal
        index = self.AnalysisTabs.currentIndex()
        tab = self.AnalysisTabs.widget(index)
        title = self.tab_editor.text()

        self.tab_editor.hide()

        self.AnalysisTabs.setTabText(index, title)
        tab.setObjectName(title)

    #
    # Workspace actions
    #
    def toggle_workspace_actions(self, index=-1):
        """
        Enable or disable workspace actions depending on
        how many tabs are open and if a policy is open.

        This is a slot for the QTabWidget.currentChanged()
        signal, though index is ignored.
        """
        open_tabs = self.AnalysisTabs.count() > 0
        open_policy = self._policy is not None

        self.log.debug("{0} actions requiring an open policy.".
                       format("Enabling" if open_policy else "Disabling"))
        self.log.debug("{0} actions requiring open tabs.".
                       format("Enabling" if open_tabs else "Disabling"))
        self.save_settings_action.setEnabled(open_tabs)
        self.save_workspace_action.setEnabled(open_tabs)
        self.new_analysis.setEnabled(open_policy)
        self.new_from_settings_action.setEnabled(open_policy)
        self.load_settings_action.setEnabled(open_tabs)

    def _get_settings(self, index=None):
        """Return a dictionary with the settings of the tab at the specified index."""
        if index is None:
            index = self.AnalysisTabs.currentIndex()

        assert index >= 0, "Tab index is negative in _get_settings.  This is an SETools bug."
        tab = self.AnalysisTabs.widget(index)

        settings = tab.save()

        # add the tab info to the settings.
        settings["__title__"] = self.AnalysisTabs.tabText(index)
        settings["__tab__"] = type(tab).__name__

        return settings

    def _put_settings(self, settings, index=None):
        """Load the settings into the specified tab."""

        if index is None:
            index = self.AnalysisTabs.currentIndex()

        assert index >= 0, "Tab index is negative in _put_settings.  This is an SETools bug."
        tab = self.AnalysisTabs.widget(index)

        if settings["__tab__"] != type(tab).__name__:
            raise TypeError("The current tab ({0}) does not match the tab in the settings file "
                            "({1}).".format(type(tab).__name__, settings["__tab__"]))

        try:
            self.AnalysisTabs.setTabText(index, str(settings["__title__"]))
        except KeyError:
            self.log.warning("Settings file does not have a title setting.")

        tab.load(settings)

    def load_settings(self, new=False):
        filename = QFileDialog.getOpenFileName(self, "Open settings file", ".",
                                               "Apol Tab Settings File (*.apolt);;"
                                               "All Files (*)")[0]
        if not filename:
            return

        try:
            with open(filename, "r") as fd:
                settings = json.load(fd)
        except ValueError as ex:
            self.log.critical("Invalid settings file \"{0}\"".format(filename))
            self.error_msg.critical(self, "Failed to load settings",
                                    "Invalid settings file: \"{0}\"".format(filename))
            return
        except OSError as ex:
            self.log.critical("Unable to load settings file \"{0.filename}\": {0.strerror}".
                              format(ex))
            self.error_msg.critical(self, "Failed to load settings",
                                    "Failed to load \"{0.filename}\": {0.strerror}".format(ex))
            return
        except Exception as ex:
            self.log.critical("Unable to load settings file \"{0}\": {1}".format(filename, ex))
            self.error_msg.critical(self, "Failed to load settings", str(ex))
            return

        self.log.info("Loading analysis settings from \"{0}\"".format(filename))

        if new:
            try:
                tabclass = TAB_REGISTRY[settings["__tab__"]]
            except KeyError:
                self.log.critical("Missing analysis type in \"{0}\"".format(filename))
                self.error_msg.critical(self, "Failed to load settings",
                                        "The type of analysis is missing in the settings file.")
                return

            # The tab title will be set by _put_settings.
            index = self.create_new_analysis("Tab", tabclass)
        else:
            index = None

        try:
            self._put_settings(settings, index)
        except Exception as ex:
            self.log.critical("Error loading settings file \"{0}\": {1}".format(filename, ex))
            self.error_msg.critical(self, "Failed to load settings",
                                    "Error loading settings file \"{0}\":\n\n{1}".
                                    format(filename, ex))
        else:
            self.log.info("Successfully loaded analysis settings from \"{0}\"".format(filename))

    def new_analysis_from_config(self):
        self.load_settings(new=True)

    def save_settings(self):
        try:
            settings = self._get_settings()

        except TabFieldError as ex:
            self.log.critical("Errors in the query prevent saving the settings. {0}".format(ex))
            self.error_msg.critical(self, "Unable to save settings",
                                    "Please resolve errors in the tab before saving the settings."
                                    )
            return

        filename = QFileDialog.getSaveFileName(self, "Save analysis tab settings", "analysis.apolt",
                                               "Apol Tab Settings File (*.apolt);;"
                                               "All Files (*)")[0]

        if not filename:
            return

        try:
            with open(filename, "w") as fd:
                json.dump(settings, fd, indent=1)
        except OSError as ex:
            self.log.critical("Unable to save settings file \"{0.filename}\": {0.strerror}".
                              format(ex))
            self.error_msg.critical(self, "Failed to save settings",
                                    "Failed to save \"{0.filename}\": {0.strerror}".format(ex))
        except Exception as ex:
            self.log.critical("Unable to save settings file \"{0}\": {1}".format(filename, ex))
            self.error_msg.critical(self, "Failed to save settings", str(ex))
        else:
            self.log.info("Successfully saved settings file \"{0}\"".format(filename))

    def load_workspace(self):
        # 1. if number of tabs > 0, check if we really want to do this
        if self.AnalysisTabs.count() > 0:
            reply = QMessageBox.question(
                self, "Continue?",
                "Loading a workspace will close all existing analyses.  Continue?",
                QMessageBox.Yes | QMessageBox.No)

            if reply == QMessageBox.No:
                return

        # 2. try to load the workspace file, if we fail, bail
        filename = QFileDialog.getOpenFileName(self, "Open workspace file", ".",
                                               "Apol Workspace Files (*.apolw);;"
                                               "All Files (*)")[0]

        if not filename:
            return

        try:
            with open(filename, "r") as fd:
                workspace = json.load(fd)
        except ValueError as ex:
            self.log.critical("Invalid workspace file \"{0}\"".format(filename))
            self.error_msg.critical(self, "Failed to load workspace",
                                    "Invalid workspace file: \"{0}\"".format(filename))
            return
        except OSError as ex:
            self.log.critical("Unable to load workspace file \"{0.filename}\": {0.strerror}".
                              format(ex))
            self.error_msg.critical(self, "Failed to load workspace",
                                    "Failed to load \"{0.filename}\": {0.strerror}".format(ex))
            return
        except Exception as ex:
            self.log.critical("Unable to load workspace file \"{0}\": {1}".format(filename, ex))
            self.error_msg.critical(self, "Failed to load workspace", str(ex))
            return

        # 3. close all tabs.  Explicitly do this to avoid the question
        #    about closing the policy with tabs open.
        self.AnalysisTabs.clear()

        # 4. close policy
        self.close_policy()

        # 5. try to open the specified policy, if we fail, bail.  Note:
        #    handling exceptions from the policy load is done inside
        #    the load_policy function, so only the KeyError needs to be caught here
        try:
            self.load_policy(workspace["__policy__"])
        except KeyError:
            self.log.critical("Missing policy in workspace file \"{0}\"".format(filename))
            self.error_msg.critical(self, "Missing policy in workspace file \"{0}\"".
                                    format(filename))

        if self._policy is None:
            self.log.critical("The policy could not be loaded in workspace file \"{0}\"".
                              format(filename))
            self.error_msg.critical(self, "The policy could not be loaded in workspace file \"{0}\""
                                    ". Aborting workspace load.".format(filename))
            return

        # 6. try to open the specified perm map, if we fail,
        #    tell the user we will continue with the default map; load the default map
        #    Note: handling exceptions from the map load is done inside
        #    the load_permmap function, so only the KeyError needs to be caught here
        try:
            self.load_permmap(workspace["__permmap__"])
        except KeyError:
            self.log.warning("Missing permission map in workspace file \"{0}\"".format(filename))
            self.error_msg.warning(self, "Missing permission map setting.",
                                   "Missing permission map in workspace file \"{0}\"".
                                   format(filename))

        if self._permmap is None:
            self.error_msg.information(self, "Loading default permission map.",
                                       "The default permisison map will be loaded.")
            self.load_permmap()

        # 7. try to open all tabs and apply settings.  Record any errors
        try:
            tab_list = list(workspace["__tabs__"])
        except KeyError:
            self.log.critical("Missing tab list in workspace file \"{0}\"".format(filename))
            self.error_msg.critical(self, "Failed to load workspace",
                                    "The workspace file is missing the tab list.  Aborting.")
            return
        except TypeError:
            self.log.critical("Invalid tab list in workspace file.")
            self.error_msg.critical(self, "Failed to load workspace",
                                    "The tab count is invalid.  Aborting.")
            return

        loading_errors = []
        for i, settings in enumerate(tab_list):
            try:
                tabclass = TAB_REGISTRY[settings["__tab__"]]
            except KeyError:
                error_str = "Missing analysis type for tab {0}. Skipping this tab.".format(i)
                self.log.error(error_str)
                loading_errors.append(error_str)
                continue

            # The tab title will be set by _put_settings.
            index = self.create_new_analysis("Tab", tabclass)

            try:
                self._put_settings(settings, index)
            except Exception as ex:
                error_str = "Error loading settings for tab {0}: {1}".format(i, ex)
                self.log.error(error_str)
                loading_errors.append(error_str)

        self.log.info("Completed loading workspace from \"{0}\"".format(filename))

        # 8. if there are any errors, open a dialog with the
        #    complete list of tab errors
        if loading_errors:
            self.error_msg.warning(self, "Errors while loading workspace:",
                                   "There were errors while loading the workspace:\n\n{0}".
                                   format("\n\n".join(loading_errors)))

    def save_workspace(self):
        workspace = {}
        save_errors = []

        workspace["__policy__"] = os.path.abspath(str(self._policy))
        workspace["__permmap__"] = os.path.abspath(str(self._permmap))
        workspace["__tabs__"] = []

        for index in range(self.AnalysisTabs.count()):
            tab = self.AnalysisTabs.widget(index)

            try:
                settings = tab.save()
            except TabFieldError as ex:
                tab_name = self.AnalysisTabs.tabText(index)
                save_errors.append(tab_name)
                self.log.error("Error: tab \"{0}\": {1}".format(tab_name, str(ex)))
            else:
                # add the tab info to the settings.
                settings["__title__"] = self.AnalysisTabs.tabText(index)
                settings["__tab__"] = type(tab).__name__

                workspace["__tabs__"].append(settings)

        if save_errors:
            self.log.critical("Errors in tabs prevent saving the workspace.")
            self.error_msg.critical(self, "Unable to save workspace",
                                    "Please resolve errors in the following tabs before saving the"
                                    " workspace:\n\n{0}".format("\n".join(save_errors)))
            return

        filename = QFileDialog.getSaveFileName(self, "Save analysis workspace", "workspace.apolw",
                                               "Apol Workspace Files (*.apolw);;"
                                               "All Files (*)")[0]

        if not filename:
            return

        with open(filename, "w") as fd:
            json.dump(workspace, fd, indent=1)

    #
    # Edit actions
    #
    def copy(self):
        """Copy text from the currently-focused widget."""
        with suppress(AttributeError):
            QApplication.instance().focusWidget().copy()

    def cut(self):
        """Cut text from the currently-focused widget."""
        with suppress(AttributeError):
            QApplication.instance().focusWidget().cut()

    def paste(self):
        """Paste text into the currently-focused widget."""
        with suppress(AttributeError):
            QApplication.instance().focusWidget().paste()

    #
    # Help actions
    #
    def about_apol(self):
        QMessageBox.about(self, "About Apol", "Version {0}<br>"
                          "Apol is a graphical SELinux policy analysis tool and part of "
                          "<a href=\"https://github.com/SELinuxProject/setools/wiki\">"
                          "SETools</a>.<p>"
                          "Copyright (C) 2015-2016, Tresys Technology<p>"
                          "Copyright (C) 2016-2019, Chris PeBenito <pebenito@ieee.org>".
                          format(__version__))

    def apol_help(self):
        """Open the main help window."""
        if self.help_process.state() != QProcess.NotRunning:
            return

        distro = pkg_resources.get_distribution("setools")
        helpfile = "{0}/setoolsgui/apol/apol.qhc".format(distro.location)

        self.log.debug("Starting assistant with help file {0}".format(helpfile))
        self.help_process.start(self.config.assistant,
                                ["-collectionFile", helpfile, "-showUrl",
                                 "qthelp://com.github.selinuxproject.setools/doc/index.html",
                                 "-show", "contents", "-enableRemoteControl"])

    @pyqtSlot(QProcess.ProcessError)
    def help_failed(self, error):
        """Starting assistant failed."""
        if error != QProcess.FailedToStart:
            return

        self.log.error("Failed to start Qt assistant {}.".format(self.config.assistant))
        self._find_assistant()

    def _find_assistant(self):
        """Try to find qt assistant in a few standard locations."""
        for name in POSSIBLE_ASSISTANT:
            self.log.debug("Trying assistant {}".format(name))
            filename = shutil.which(name, mode=os.F_OK | os.X_OK, path=BIN_SEARCH_PATHS)
            if filename:
                self.log.debug("Assistant {} is available and executable.".format(filename))
                break
        else:
            reply = QMessageBox.question(
                self, "Qt Assistant Package Installed?",
                "Failed to start QT Assistant program {}. "
                "This is typically in the assistant or qt5-assistant package. "
                "Choose location of Qt Assistant executable?".format(
                    self.config.assistant),
                QMessageBox.Yes | QMessageBox.No)

            if reply == QMessageBox.No:
                return

            filename = QFileDialog.getOpenFileName(self, "Location of qt-assistant executable",
                                                   "/usr/bin",
                                                   "All Files (*)")[0]

            self.log.debug("User chose assistant {}.".format(filename))

        if filename:
            self.log.debug("Updating config with assistant {}".format(filename))
            self.config.assistant = filename
            self.config.save()
            self.apol_help()

    @pyqtSlot(str)
    def set_help(self, location):
        """Set the help window to the specified document."""
        if self.help_process.state() == QProcess.NotStarted:
            self.apol_help()
            if not self.help_process.waitForStarted():
                self.log.warning("Timed out waiting for Qt assistant to start.")
                return
        elif self.help_process.state() == QProcess.Starting:
            if not self.help_process.waitForStarted():
                self.log.warning("Timed out waiting for Qt assistant to start.")
                return

        self.help_process.write("setSource qthelp://com.github.selinuxproject.setools/doc/{0}\n".
                                format(location))



# Copyright 2016, Tresys Technology, LLC
#
# This file is part of SETools.
#
# SETools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1 of
# the License, or (at your option) any later version.
#
# SETools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with SETools.  If not, see
# <http://www.gnu.org/licenses/>.
#
import logging

from PyQt5.QtGui import QFont, QTextCursor
from PyQt5.QtWidgets import QDialog

from .widget import SEToolsWidget


class DetailsPopup(SEToolsWidget, QDialog):

    """A generic non-modal popup with a text field to write detailed info."""
    # TODO: make the font changes relative
    # instead of setting absolute values

    def __init__(self, parent, title=None):
        super(DetailsPopup, self).__init__(parent)
        self.log = logging.getLogger(__name__)
        self.setupUi(title)

    def setupUi(self, title):
        self.load_ui("detail_popup.ui")

        if title:
            self.title = title

    @property
    def title(self):
        self.windowTitle(self)

    @title.setter
    def title(self, text):
        self.setWindowTitle(text)

    def append(self, text):
        self.contents.setFontWeight(QFont.Normal)
        self.contents.setFontPointSize(9)
        self.contents.append(text)

    def append_header(self, text):
        self.contents.setFontWeight(QFont.Black)
        self.contents.setFontPointSize(11)
        self.contents.append(text)

    def show(self):
        self.contents.moveCursor(QTextCursor.Start)
        super(DetailsPopup, self).show()



# Copyright 2014-2015, Tresys Technology, LLC
#
# This file is part of SETools.
#
# SETools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1 of
# the License, or (at your option) any later version.
#
# SETools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with SETools.  If not, see
# <http://www.gnu.org/licenses/>.
#
import logging
import re
from contextlib import suppress
from typing import Iterable

from .descriptors import CriteriaDescriptor, CriteriaSetDescriptor
from .exception import NoCommon
from .mixins import MatchName
from .policyrep import ObjClass
from .query import PolicyQuery
from .util import match_regex, match_regex_or_set


class ObjClassQuery(MatchName, PolicyQuery):

    """
    Query object classes.

    Parameter:
    policy          The policy to query.

    Keyword Parameters/Class attributes:
    name            The name of the object set to match.
    name_regex      If true, regular expression matching will
                    be used for matching the name.
    common          The name of the inherited common to match.
    common_regex    If true, regular expression matching will
                    be used for matching the common name.
    perms           The permissions to match.
    perms_equal     If true, only commons with permission sets
                    that are equal to the criteria will
                    match.  Otherwise, any intersection
                    will match.
    perms_regex     If true, regular expression matching
                    will be used on the permission names instead
                    of set logic.
                    comparison will not be used.
    perms_indirect  If false, permissions inherited from a common
                    permission set not will be evaluated.  Default
                    is true.
    """

    common = CriteriaDescriptor("common_regex", "lookup_common")
    common_regex: bool = False
    perms = CriteriaSetDescriptor("perms_regex")
    perms_equal: bool = False
    perms_indirect: bool = True
    perms_regex: bool = False

    def __init__(self, policy, **kwargs) -> None:
        super(ObjClassQuery, self).__init__(policy, **kwargs)
        self.log = logging.getLogger(__name__)

    def results(self) -> Iterable[ObjClass]:
        """Generator which yields all matching object classes."""
        self.log.info("Generating object class results from {0.policy}".format(self))
        self._match_name_debug(self.log)
        self.log.debug("Common: {0.common!r}, regex: {0.common_regex}".format(self))
        self.log.debug("Perms: {0.perms}, regex: {0.perms_regex}, "
                       "eq: {0.perms_equal}, indirect: {0.perms_indirect}".format(self))

        for class_ in self.policy.classes():
            if not self._match_name(class_):
                continue

            if self.common:
                try:
                    if not match_regex(
                            class_.common,
                            self.common,
                            self.common_regex):
                        continue
                except NoCommon:
                    continue

            if self.perms:
                perms = class_.perms

                if self.perms_indirect:
                    with suppress(NoCommon):
                        perms |= class_.common.perms

                if not match_regex_or_set(
                        perms,
                        self.perms,
                        self.perms_equal,
                        self.perms_regex):
                    continue

            yield class_



# Copyright 2014-2015, Tresys Technology, LLC
#
# This file is part of SETools.
#
# SETools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1 of
# the License, or (at your option) any later version.
#
# SETools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with SETools.  If not, see
# <http://www.gnu.org/licenses/>.
#
import logging
import re
from typing import cast, Iterable

from .query import PolicyQuery
from .descriptors import CriteriaDescriptor, CriteriaSetDescriptor
from .mixins import MatchObjClass
from .policyrep import AnyDefault, DefaultRange, DefaultRuletype, DefaultValue, DefaultRangeValue


class DefaultQuery(MatchObjClass, PolicyQuery):

    """
    Query default_* statements.

    Parameter:
    policy          The policy to query.

    Keyword Parameters/Class attributes:
    ruletype        The rule type(s) to match.
    tclass          The object class(es) to match.
    tclass_regex    If true, use a regular expression for
                    matching the rule's object class.
    default         The default to base new contexts (e.g. "source" or "target")
    default_range   The range to use on new context, default_range only
                    ("low", "high", "low_high")
    """

    ruletype = CriteriaSetDescriptor(enum_class=DefaultRuletype)
    default = CriteriaDescriptor(enum_class=DefaultValue)
    default_range = CriteriaDescriptor(enum_class=DefaultRangeValue)

    def __init__(self, policy, **kwargs) -> None:
        super(DefaultQuery, self).__init__(policy, **kwargs)
        self.log = logging.getLogger(__name__)

    def results(self) -> Iterable[AnyDefault]:
        """Generator which yields all matching default_* statements."""
        self.log.info("Generating default_* results from {0.policy}".format(self))
        self.log.debug("Ruletypes: {0.ruletype!r}".format(self))
        self._match_object_class_debug(self.log)
        self.log.debug("Default: {0.default!r}".format(self))
        self.log.debug("Range: {0.default_range!r}".format(self))

        for d in self.policy.defaults():
            if self.ruletype and d.ruletype not in self.ruletype:
                continue

            if not self._match_object_class(d):
                continue

            if self.default and d.default != self.default:
                continue

            if self.default_range:
                try:
                    if cast(DefaultRange, d).default_range != self.default_range:
                        continue
                except AttributeError:
                    continue

            yield d



# Copyright 2020, Microsoft Corporation
#
# This file is part of SETools.
#
# SETools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1 of
# the License, or (at your option) any later version.
#
# SETools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with SETools.  If not, see
# <http://www.gnu.org/licenses/>.
#

import sys
import configparser
import logging
from datetime import datetime, timezone
from typing import List

from ..exception import InvalidCheckerConfig, InvalidCheckerModule
from ..policyrep import SELinuxPolicy

from .checkermodule import CHECKER_REGISTRY, CheckerModule
from .globalkeys import CHECK_TYPE_KEY


SECTION_SEPARATOR = "---------------------------------------------------------\n\n"


class PolicyChecker:

    """Configuration file-driven automated policy analysis checks."""

    def __init__(self, policy: SELinuxPolicy, configpath: str) -> None:
        assert CHECKER_REGISTRY, "No checks are loaded, this is a bug."

        self.log = logging.getLogger(__name__)
        self.policy = policy
        self.checks: List[CheckerModule] = []
        self.config = configpath

    @property
    def config(self):
        return self._configpath

    @config.setter
    def config(self, configpath: str):
        self.log.info("Opening policy checker config {}.".format(configpath))
        try:
            with open(configpath, "r") as fd:
                config = configparser.ConfigParser()
                config.read_file(fd, source=configpath)
        except Exception as e:
            raise InvalidCheckerConfig("Unable to parse checker config {}: {}".format(
                configpath, e)) from e

        self.log.info("Validating configuration settings.")

        checks = []
        for checkname, checkconfig in config.items():
            if checkname == "DEFAULT":
                # top level/DEFAULT section is not a check.
                continue

            try:
                check_type = checkconfig[CHECK_TYPE_KEY]
            except KeyError as e:
                raise InvalidCheckerModule("{}: Missing {} option.".
                                           format(checkname, CHECK_TYPE_KEY))

            try:
                newcheck = CHECKER_REGISTRY[check_type](self.policy, checkname, checkconfig)
            except KeyError as e:
                raise InvalidCheckerModule("{}: Unknown policy check type: {}".
                                           format(checkname, check_type)) from e

            checks.append(newcheck)

        if not checks:
            raise InvalidCheckerConfig("No checks found in {}.".format(configpath))

        self.log.debug("Validated {} checks.".format(len(self.checks)))
        self.log.info("Successfully opened policy checker config {}.".format(configpath))
        self._configpath = configpath
        self.checks = checks
        self._config = config

    def run(self, output=sys.stdout) -> int:
        """Run all configured checks and print report to the file-like output."""
        failures = 0

        assert self.checks, "Configuration loaded but no checks configured. This is a bug."

        output.write(SECTION_SEPARATOR)
        output.write("Policy check configuration: {}\n".format(self.config))
        output.write("Policy being checked: {}\n".format(self.policy))
        output.write("Start time: {}\n\n".format(datetime.now(timezone.utc)))

        result_summary = []
        for check in self.checks:

            check_failures = 0
            try:
                output.write(SECTION_SEPARATOR)
                output.write("Check name: {}\n\n".format(check.checkname))
                if check.desc:
                    output.write("Description: {}\n\n".format(check.desc))

                if check.disable:
                    output.write("Check DISABLED.  Reason: {}\n\n".format(check.disable))
                    result_summary.append((check.checkname, "DISABLED ({})".format(check.disable)))
                    self.log.debug("Skipping disabled check {!r}: {}".format(check.checkname,
                                   check.disable))
                    continue

                self.log.debug("Running check {0!r}, type {1}.".format(
                    check.checkname, check.check_type))
                check.output = output
                check_failures += len(check.run())
                output.write("\n")
            except Exception as e:
                output.write("Unexpected error: {}.  Failing check.\n\n".format(e))
                self.log.debug("Exception info", exc_info=e)
                check_failures += 1

            if check_failures:
                output.write("Check FAILED\n\n")
                result_summary.append((check.checkname, "FAILED ({} failures)".format(
                    check_failures)))
            else:
                output.write("Check PASSED\n\n")
                result_summary.append((check.checkname, "PASSED"))

            failures += check_failures

        output.write(SECTION_SEPARATOR)
        output.write("Result Summary:\n\n")
        for checkname, result in result_summary:
            output.write("{:<39} {}\n".format(checkname, result))

        output.write("\n{} failure(s) found.\n\n".format(failures))
        output.write("Policy check configuration: {}\n".format(self.config))
        output.write("Policy being checked: {}\n".format(self.policy))
        output.write("End time: {}\n".format(datetime.now(timezone.utc)))
        self.log.info("{} failures found in {} checks.".format(failures, len(self.checks)))
        return failures
